"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from functools import partial
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union, Hashable
import traceback

import numpy as np

from fuse.eval.metrics.metrics_common import MetricDefault, MetricWithCollectorBase
from fuse.eval.metrics.libs.classification import MetricsLibClass


class MetricMultiClassDefault(MetricWithCollectorBase):
    """
    Default generic implementation for metric
    Can be used for any metric getting as an input list of prediction, list of targets and optionally additional parameters
    """

    def __init__(
        self,
        pred: str,
        target: str,
        metric_func: Callable,
        class_names: Optional[Sequence[str]] = None,
        class_weights: Optional[Sequence[float]] = None,
        **kwargs,
    ):
        """
        :param pred: prediction key to collect
        :param target: target key to collect
        :param metric_func: function getting as a input list of predictions, targets and optionally more arguments specified in kwargs
                            the function should return a result or a dictionary of results
        :param class_names: class names for multi-class evaluation or None for binary evaluation
        :param class_weight: weight per class - the macro_average result will be a weighted sum rather than an average
        :param kwargs: additional kw arguments for MetricWithCollectorBase
        """
        super().__init__(pred=pred, target=target, **kwargs)
        self._metric_func = metric_func
        self._class_names = class_names
        self._class_weights = class_weights

    def eval(
        self, results: Dict[str, Any] = None, ids: Optional[Sequence[Hashable]] = None
    ) -> Union[Dict[str, Any], Any]:
        """
        See super class
        """
        # extract values from collected data and results dict
        kwargs = self._extract_arguments(results, ids)

        if self._class_names is None:
            # single evaluation for all classes at once / binary classifier
            try:
                metric_results = self._metric_func(**kwargs)
            except:
                track = traceback.format_exc()
                print(f"Error in metric: {track}")
                metric_results = None
        else:
            metric_results = {}

            # one vs rest evaluation per class, including average
            try:
                # compute one-vs-rest metrics
                is_dict = False
                all_classes = []
                for cls_index, cls_name in enumerate(self._class_names):
                    cls_res = self._metric_func(pos_class_index=cls_index, **kwargs)
                    if isinstance(cls_res, dict):
                        is_dict = True
                        for sub_metric_name in cls_res:
                            metric_results[f"{sub_metric_name}.{cls_name}"] = cls_res[sub_metric_name]
                    else:
                        assert (
                            is_dict is False
                        ), "expect all sub metric results to either return dictionary or single value"
                        metric_results[f"{cls_name}"] = cls_res

                    all_classes.append(cls_res)

                # compute macro average
                if is_dict:
                    for key in all_classes[0]:
                        all_classes_elem = np.array([d[key] for d in all_classes])
                        indices = ~np.isnan(all_classes_elem)
                        if self._class_weights is None:
                            weights = None
                        else:
                            weights = self._class_weights[indices]
                        metric_results[f"{key}.macro_avg"] = np.average(all_classes_elem[indices], weights=weights)
                else:
                    all_classes = np.array(all_classes)
                    indices = ~np.isnan(all_classes)
                    if self._class_weights is None:
                        weights = None
                    else:
                        weights = self._sum_weights[indices]
                    metric_results["macro_avg"] = np.average(all_classes[indices], weights=weights)
            except:
                track = traceback.format_exc()
                print(f"Error in metric: {type(self).__name__} - {track}")
                for cls_name in self._class_names:
                    metric_results[f"{cls_name}"] = None
                metric_results["macro_avg"] = None

        return metric_results


class MetricAUCROC(MetricMultiClassDefault):
    """
    Compute auc roc (Receiver operating characteristic) score using sklearn (one vs rest)
    """

    def __init__(
        self,
        pred: str,
        target: str,
        class_names: Optional[Sequence[str]] = None,
        max_fpr: Optional[float] = None,
        **kwargs,
    ):
        """
        See MetricMultiClassDefault for the missing params
        :param max_fpr: float > 0 and <= 1, default=None
                        If not ``None``, the standardized partial AUC over the range [0, max_fpr] is returned.
        """
        auc_roc = partial(MetricsLibClass.auc_roc, max_fpr=max_fpr)
        super().__init__(pred, target, metric_func=auc_roc, class_names=class_names, **kwargs)


class MetricROCCurve(MetricDefault):
    """
    Output the roc curve (Multi class version - one vs rest)
    """

    def __init__(
        self,
        pred: str,
        target: str,
        output_filename: Optional[str] = None,
        class_names: Optional[List[str]] = None,
        sample_weight: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        :param pred:                key for predicted output (e.g., class scores after softmax)
        :param target:              key for target (e.g., ground truth label)
        :param output_filename:     output filename to save the figure. if None, will just save the arguments to create the figure in results dictionary.
        :param class_names:         Name for each class otherwise will assume binary classification and create curve just for class 1.
        :param sample_weight:       Optional, key for  sample weight
        :param kwargs:              super class arguments
        """
        roc = partial(MetricsLibClass.roc_curve, class_names=class_names, output_filename=output_filename)
        super().__init__(pred=pred, target=target, sample_weight=sample_weight, metric_func=roc, **kwargs)


class MetricAUCPR(MetricMultiClassDefault):
    """
    Compute Area Under Precision Recall Curve score using sklearn (one vs rest)
    """

    def __init__(self, pred: str, target: str, class_names: Optional[Sequence[str]] = None, **kwargs):
        super().__init__(pred, target, MetricsLibClass.auc_pr, class_names=class_names, **kwargs)


class MetricAccuracy(MetricDefault):
    """
    Compute accuracy over all the samples
    """

    def __init__(self, pred: str, target: str, sample_weight: Optional[str] = None, **kwargs):
        """
        See MetricDefault for the missing params
        :param sample_weight: weight per sample for the final accuracy score. Keep None if not required.
        """
        super().__init__(
            pred=pred, target=target, sample_weight=sample_weight, metric_func=MetricsLibClass.accuracy, **kwargs
        )


class MetricConfusion(MetricMultiClassDefault):
    """
    Compute metrics derived from one-vs-rest confusion matrix such as 'sensitivity', 'recall', 'tpr', 'specificity',  'selectivity', 'npr', 'precision', 'ppv', 'f1'
    """

    def __init__(
        self,
        pred: str,
        target: str,
        class_names: Optional[Sequence[str]] = None,
        metrics: Sequence[str] = ("sensitivity",),
        operation_point: Union[float, Sequence[Tuple[int, float]], str, None] = tuple(),
        **kwargs,
    ):
        """
        See MetricMultiClassDefault for the missing params
        :param metrics: sequence of required metrics that dervied from confusion matrix.
                        Options: 'sensitivity', 'recall', 'tpr', 'specificity',  'selectivity', 'npr', 'precision', 'ppv', 'f1'
        """
        super().__init__(
            pred=pred,
            target=target,
            metric_func=MetricsLibClass.confusion_metrics,
            class_names=class_names,
            metrics=metrics,
            **kwargs,
        )
        self._metrics = metrics


class MetricConfusionMatrix(MetricDefault):
    """
    Computes the confusion matrix from the data
    Multi class version
    returns dataframe
    """

    def __init__(
        self, cls_pred: str, target: str, class_names: Sequence[str], sample_weight: Optional[str] = None, **kwargs
    ) -> None:
        """
        See super class
        :param cls_pred: key to class predictions
        """
        conf_matrix_func = partial(MetricsLibClass.confusion_matrix, class_names=class_names)
        super().__init__(
            pred=None,
            cls_pred=cls_pred,
            target=target,
            sample_weight=sample_weight,
            metric_func=conf_matrix_func,
            **kwargs,
        )


class MetricBSS(MetricDefault):
    """
    Multi class version for Brier Skill Score (BSS)
    bss = 1 - bs / bs_{ref}

    bs_{ref} will be computed for a model that makes a predictions according to the prevalance of each class in dataset

    """

    def __init__(self, pred: str, target: str, **kwargs):
        """
        See super class
        """
        super().__init__(pred=pred, target=target, metric_func=MetricsLibClass.multi_class_bss, **kwargs)



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021
"""

import unittest

from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseUtilsHierarchicalDictTestCase(unittest.TestCase):
    def setUp(self):
        self.hierarchical_dict = {"a": 1, "b": {"c": 2, "d": 3}, "c": 4}

    def test_get(self):
        self.assertEqual(FuseUtilsHierarchicalDict.get(self.hierarchical_dict, "a"), 1)
        self.assertEqual(FuseUtilsHierarchicalDict.get(self.hierarchical_dict, "b.c"), 2)
        self.assertEqual(FuseUtilsHierarchicalDict.get(self.hierarchical_dict, "c"), 4)
        self.assertDictEqual(FuseUtilsHierarchicalDict.get(self.hierarchical_dict, "b"), {"c": 2, "d": 3})

    def test_set(self):
        FuseUtilsHierarchicalDict.set(self.hierarchical_dict, "a", 7)
        self.assertEqual(self.hierarchical_dict["a"], 7)
        FuseUtilsHierarchicalDict.set(self.hierarchical_dict, "a", 1)
        self.assertEqual(self.hierarchical_dict["a"], 1)

        FuseUtilsHierarchicalDict.set(self.hierarchical_dict, "b.d", 5)
        self.assertEqual(self.hierarchical_dict["b"]["d"], 5)
        FuseUtilsHierarchicalDict.set(self.hierarchical_dict, "b.d", 3)
        self.assertEqual(self.hierarchical_dict["b"]["d"], 3)

        FuseUtilsHierarchicalDict.set(self.hierarchical_dict, "b.e.f.g", 9)
        self.assertEqual(self.hierarchical_dict["b"]["e"]["f"]["g"], 9)
        del self.hierarchical_dict["b"]["e"]

    def test_pop(self):
        FuseUtilsHierarchicalDict.pop(self.hierarchical_dict, "b.d")
        all_keys = FuseUtilsHierarchicalDict.get_all_keys(self.hierarchical_dict)
        self.assertSetEqual(set(all_keys), {"a", "b.c", "c"})

        FuseUtilsHierarchicalDict.pop(self.hierarchical_dict, "c")
        all_keys = FuseUtilsHierarchicalDict.get_all_keys(self.hierarchical_dict)
        self.assertSetEqual(set(all_keys), {"a", "b.c"})
        with self.assertRaises(KeyError):
            FuseUtilsHierarchicalDict.pop(self.hierarchical_dict, "c")
        with self.assertRaises(KeyError):
            FuseUtilsHierarchicalDict.pop(self.hierarchical_dict, "lala")

    def test_get_all_keys(self):
        all_keys = FuseUtilsHierarchicalDict.get_all_keys(self.hierarchical_dict)
        self.assertSetEqual(set(all_keys), {"a", "b.c", "b.d", "c"})

    def test_is_in(self):
        self.assertTrue(FuseUtilsHierarchicalDict.is_in(self.hierarchical_dict, "a"))
        self.assertTrue(FuseUtilsHierarchicalDict.is_in(self.hierarchical_dict, "b.c"))
        self.assertTrue(FuseUtilsHierarchicalDict.is_in(self.hierarchical_dict, "b.d"))
        self.assertTrue(FuseUtilsHierarchicalDict.is_in(self.hierarchical_dict, "c"))
        self.assertFalse(FuseUtilsHierarchicalDict.is_in(self.hierarchical_dict, "d"))
        self.assertFalse(FuseUtilsHierarchicalDict.is_in(self.hierarchical_dict, "e"))

    def tearDown(self):
        delattr(self, "hierarchical_dict")


if __name__ == "__main__":
    unittest.main()



import unittest

from fuse.utils.data import CollateToBatchList


class TestTimer(unittest.TestCase):
    def test_collate(self):
        """ """

        x = [{"a": 1, "b": 2}, {"a": 3, "b": 4}, {"a": 5, "b": 6}]
        ref = {"a": [1, 3, 5], "b": [2, 4, 6]}
        col = CollateToBatchList(x)
        y = col(x)
        self.assertEqual(y.to_dict(), ref)


if __name__ == "__main__":
    unittest.main()



import pathlib
import sys
import os
from typing import List, Tuple, Union
from collections import OrderedDict

import csv
from fuse.eval.metrics.classification.metrics_thresholding_common import MetricApplyThresholds
from fuse.utils.ndict import NDict
import pandas as pd
import numpy as np

from fuse.utils.file_io.file_io import create_or_reset_dir

from fuse.eval.evaluator import EvaluatorDefault
from fuse.eval.metrics.classification.metrics_classification_common import MetricBSS, MetricConfusion
from fuse.eval.metrics.metrics_common import CI


## Constants
# Constants that defines the expected format of the prediction and target files and list the classes for task 1 and task @
EXPECTED_TASK1_PRED_KEYS = {
    "image_name",
    "predicted_label",
    "Noncancerous-score",
    "Precancerous-score",
    "Cancerous-score",
}
EXPECTED_TASK2_PRED_KEYS = {
    "image_name",
    "predicted_label",
    "PB-score",
    "UDH-score",
    "FEA-score",
    "ADH-score",
    "DCIS-score",
    "IC-score",
}
EXPECTED_TARGET_KEYS = {"image_name", "Task1-target", "Task2-target"}
PRED_SAMPLE_DESC_NAME = "image_name"
TARGET_SAMPLE_DESC_NAME = "image_name"
TASK1_CLASS_NAMES = ("Noncancerous", "Precancerous", "Cancerous")  # must be aligned with task1 targets
TASK2_CLASS_NAMES = ("PB", "UDH", "FEA", "ADH", "DCIS", "IC")  # must be aligned with task2 targets


def process(sample_dict: NDict) -> dict:
    """
    post caching processing. Will group together to an array the per class scores and verify it sums up to 1.0
    :param sample_dict: a dictionary that contais all the extracted values of a single sample
    :return: a modified/alternative dictionary
    """
    # verify sample
    expected_keys = [f"task1_pred.{key}" for key in EXPECTED_TASK1_PRED_KEYS]
    expected_keys += [f"task2_pred.{key}" for key in EXPECTED_TASK2_PRED_KEYS]
    expected_keys += [f"target.{key}" for key in EXPECTED_TARGET_KEYS]
    set(expected_keys).issubset(set(sample_dict.keypaths()))

    # convert scores to numpy array
    # task 1
    task1_pred = []
    for cls_name in TASK1_CLASS_NAMES:
        task1_pred.append(sample_dict[f"task1_pred.{cls_name}-score"])
    task1_pred_array = np.array(task1_pred)
    if not np.isclose(task1_pred_array.sum(), 1.0, rtol=0.05):
        raise Exception(
            f"Error: expecting task 1 prediction for case {sample_dict['descriptor']} to sum up to almost 1.0, got {task1_pred_array}"
        )
    sample_dict["task1_pred.array"] = task1_pred_array

    # task 2
    task2_pred = []
    for cls_name in TASK2_CLASS_NAMES:
        task2_pred.append(sample_dict[f"task2_pred.{cls_name}-score"])
    task2_pred_array = np.array(task2_pred)
    if not np.isclose(task2_pred_array.sum(), 1.0, rtol=0.05):
        raise Exception(
            f"Error: expecting task 2 prediction for case {sample_dict['descriptor']} to sum up to almost 1.0, got {task2_pred_array}"
        )
    sample_dict["task2_pred.array"] = task2_pred_array

    return sample_dict


def decode_results(results: dict, output_dir: str) -> Tuple[OrderedDict, str]:
    """
    Gets the results computed by the dictionary and summarize it in a markdown text and dictionary.
    The dictionary will be saved in <output_dir>/results.csv and the markdown text in <output_dir>/results.md
    :param results: the results computed by the metrics
    :param output_dir: path to an output directory
    :return: ordered dict summarizing the results and markdown text
    """
    results = NDict(results["metrics"])
    results_table = OrderedDict()
    # Table
    ## task1
    results_table["Task1-F1"] = f"{results['task1_f1.f1.macro_avg.org']:.3f}"
    results_table[
        "Task1-F1-CI"
    ] = f"[{results['task1_f1.f1.macro_avg.conf_lower']:.3f}-{results['task1_f1.f1.macro_avg.conf_upper']:.3f}]"
    for cls_name in TASK1_CLASS_NAMES:
        results_table[f"Task1-F1-{cls_name}VsRest"] = f"{results[f'task1_f1.f1.{cls_name}.org']:.3f}"
        results_table[
            f"Task1-F1-{cls_name}VsRest-CI"
        ] = f"[{results[f'task1_f1.f1.{cls_name}.conf_lower']:.3f}-{results[f'task1_f1.f1.{cls_name}.conf_upper']:.3f}]"
    results_table["Task1-BSS"] = f"{results['task1_bss.org']:.3f}"
    results_table["Task1-BSS-CI"] = f"[{results['task1_bss.conf_lower']:.3f}-{results['task1_bss.conf_upper']:.3f}]"

    ## task2
    results_table["Task2-F1"] = f"{results['task2_f1.f1.macro_avg.org']:.3f}"
    results_table[
        "Task2-F1-CI"
    ] = f"[{results['task2_f1.f1.macro_avg.conf_lower']:.3f}-{results['task2_f1.f1.macro_avg.conf_upper']:.3f}]"
    for cls_name in TASK2_CLASS_NAMES:
        results_table[f"Task2-F1-{cls_name}VsRest"] = f"{results[f'task2_f1.f1.{cls_name}.org']:.3f}"
        results_table[
            f"Task2-F1-{cls_name}VsRest-CI"
        ] = f"[{results[f'task2_f1.f1.{cls_name}.conf_lower']:.3f}-{results[f'task2_f1.f1.{cls_name}.conf_upper']:.3f}]"
    results_table["Task2-BSS"] = f"{results['task2_bss.org']:.3f}"
    results_table["Task2-BSS-CI"] = f"[{results['task2_bss.conf_lower']:.3f}-{results['task2_bss.conf_upper']:.3f}]"

    # mark down text
    results_text = ""
    ## task 1
    results_text += "# Task 1 - 3-class WSI classification\n"
    results_text += f"F1: {results_table['Task1-F1']} {results_table['Task1-F1-CI']}\n"

    results_text += "## Multi-Class F1\n"
    table_columns = ["F1"] + [f"F1-{cls_name}VsRest" for cls_name in TASK1_CLASS_NAMES]
    results_text += "\n|"
    results_text += "".join([f" {column} |" for column in table_columns])
    results_text += "\n|"
    results_text += "".join([" ------ |" for column in table_columns])
    results_text += "\n|"
    results_text += "".join(
        [f" {results_table[f'Task1-{column}']} {results_table[f'Task1-{column}-CI']} |" for column in table_columns]
    )

    results_text += "\n## BSS\n"
    results_text += f"BSS: {results_table['Task1-BSS']} {results_table['Task1-BSS-CI']}\n"

    ## task 2
    results_text += "\n# Task 2 - 6-class WSI classification\n"
    results_text += f"F1: {results_table['Task2-F1']} {results_table['Task2-F1-CI']}\n"

    results_text += "## Multi-Class F1\n"
    table_columns = ["F1"] + [f"F1-{cls_name}VsRest" for cls_name in TASK2_CLASS_NAMES]
    results_text += "\n|"
    results_text += "".join([f" {column} |" for column in table_columns])
    results_text += "\n|"
    results_text += "".join([" ------ |" for column in table_columns])
    results_text += "\n|"
    results_text += "".join(
        [f" {results_table[f'Task2-{column}']} {results_table[f'Task2-{column}-CI']} |" for column in table_columns]
    )

    results_text += "\n## BSS\n"
    results_text += f"BSS: {results_table['Task2-BSS']} {results_table['Task2-BSS-CI']}\n"

    # save files
    with open(os.path.join(output_dir, "results.md"), "w") as output_file:
        output_file.write(results_text)

    with open(os.path.join(output_dir, "results.csv"), "w") as output_file:
        w = csv.writer(output_file)
        w.writerow(results_table.keys())
        w.writerow(results_table.values())

    return results_table, results_text


def eval(
    task1_prediction_filename: str,
    task2_prediction_filename: str,
    target_filename: str,
    output_dir: str,
    samples_descr_source: Union[str, List[str]] = "task1_pred",
) -> Tuple[OrderedDict, str]:
    """
    Load the prediction and target files, evaluate the predictions and save the results into files.
    :param task1_prediction_filename: path to a prediction csv file for task1. Expecting the columns listed in EXPECTED_TASK1_PRED_KEYS to exist (including the header)
    :param task2_prediction_filename: path to a prediction csv file for task2. Expecting the columns listed in EXPECTED_TASK2_PRED_KEYS to exist (including the header)
    :param target_filename: path to a prediction csv file. Expecting the columns listed in EXPECTED_TARGET_KEYS to exist
    :param output_dir: path to directory to save the output files
    :param samples_descr_source: will run the evaluation on the specified list of sample descriptors. Supported values:
                                 * "task1_pred" to evaluate all the samples specified in task1 prediction file.
                                 * "task2_pred" to evaluate all the samples specified in task2 prediction file.
                                 * "target" to evaluate all the samples specified in targets file.
                                 * list to define the samples explicitly
    :return: ordered dict summarizing the results and markdown text
    """
    create_or_reset_dir(output_dir, force_reset=True)

    # metrics to evaluate
    metrics = OrderedDict(
        [
            # task 1
            ("task1_op", MetricApplyThresholds(pred="task1_pred.predicted_label")),  # will apply argmax
            (
                "task1_f1",
                CI(
                    MetricConfusion(
                        pred="results:metrics.task1_op.cls_pred",
                        target="target.Task1-target",
                        class_names=TASK1_CLASS_NAMES,
                        metrics=("f1",),
                        pre_collect_process_func=process,
                    ),
                    stratum="target.Task1-target",
                ),
            ),
            (
                "task1_bss",
                CI(
                    MetricBSS(pred="task1_pred.array", target="target.Task1-target", pre_collect_process_func=process),
                    stratum="target.Task1-target",
                ),
            ),
            # task 2
            ("task2_op", MetricApplyThresholds(pred="task2_pred.predicted_label")),  # will apply argmax
            (
                "task2_f1",
                CI(
                    MetricConfusion(
                        pred="results:metrics.task1_op.cls_pred",
                        target="target.Task2-target",
                        class_names=TASK2_CLASS_NAMES,
                        metrics=("f1",),
                        pre_collect_process_func=process,
                    ),
                    stratum="target.Task2-target",
                ),
            ),
            (
                "task2_bss",
                CI(
                    MetricBSS(pred="task2_pred.array", target="target.Task2-target", pre_collect_process_func=process),
                    stratum="target.Task2-target",
                ),
            ),
        ]
    )
    # read files
    task1_pred_df = pd.read_csv(task1_prediction_filename, dtype={PRED_SAMPLE_DESC_NAME: object})
    task2_pred_df = pd.read_csv(task2_prediction_filename, dtype={PRED_SAMPLE_DESC_NAME: object})
    target_df = pd.read_csv(target_filename, dtype={TARGET_SAMPLE_DESC_NAME: object})

    # verify input
    assert set(task1_pred_df.keys()).issubset(
        EXPECTED_TASK1_PRED_KEYS
    ), f"Expecting task1 prediction file {os.path.abspath(task1_prediction_filename)} to include also the following keys: {EXPECTED_TASK1_PRED_KEYS - set(task1_pred_df.keys())}"
    assert set(task2_pred_df.keys()).issubset(
        EXPECTED_TASK2_PRED_KEYS
    ), f"Expecting task2 prediction file {os.path.abspath(task2_prediction_filename)} to include also the following keys: {EXPECTED_TASK2_PRED_KEYS - set(task2_pred_df.keys())}"
    assert set(target_df.keys()).issubset(
        EXPECTED_TARGET_KEYS
    ), f"Expecting target file {os.path.abspath(target_filename)} to include also the following keys: {EXPECTED_TARGET_KEYS - set(target_df.keys())}"

    # analyze
    task1_pred_df["id"] = task1_pred_df["image_name"]
    task2_pred_df["id"] = task2_pred_df["image_name"]
    target_df["id"] = target_df["image_name"]
    evaluator = EvaluatorDefault()
    results = evaluator.eval(
        ids=list(target_df["id"]),
        data={"task1_pred": task1_pred_df, "task2_pred": task2_pred_df, "target": target_df},
        metrics=metrics,
        output_dir=None,
    )

    # output
    return decode_results(results, output_dir=output_dir)


if __name__ == "__main__":
    """
    Run evaluation:
    Usage: python eval.py <target_filename> <task1_prediction_filename> <task2_prediction_filename> <output dir>
    Run dummy example (set the working dir to fuse-med-ml/examples/fuse_examples/imaging/classification/bright/eval): python eval.py example/example_targets.csv example/example_task1_predictions.csv example/example_task2_predictions.csv example/results
    Run baseline (set the working dir to fuse-med-ml/examples/fuse_examples/imaging/classification/bright/eval): python eval.py validation_targets.csv baseline/validation_baseline_task1_predictions.csv baseline/validation_baseline_task2_predictions.csv baseline/validation_results
    """

    if len(sys.argv) == 1:
        dir_path = pathlib.Path(__file__).parent.resolve()
        target_filename = os.path.join(dir_path, "validation_targets.csv")
        task1_prediction_filename = os.path.join(dir_path, "baseline/validation_baseline_task1_predictions.csv")
        task2_prediction_filename = os.path.join(dir_path, "baseline/validation_baseline_task2_predictions.csv")
        output_dir = os.path.join(dir_path, "baseline/validation_results")
        eval(
            target_filename=target_filename,
            task1_prediction_filename=task1_prediction_filename,
            task2_prediction_filename=task2_prediction_filename,
            output_dir=output_dir,
        )
    else:
        assert (
            len(sys.argv) == 5
        ), f"Error: expecting 4 input arguments, but got {len(sys.argv)-1}. Usage: python eval.py <target_filename> <task1_prediction_filename> <task2_prediction_filename> <output dir>"
        eval(
            target_filename=sys.argv[1],
            task1_prediction_filename=sys.argv[2],
            task2_prediction_filename=sys.argv[3],
            output_dir=sys.argv[4],
        )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import unittest
from fuse.utils.rand.seed import Seed
from fuse.utils.ndict import NDict

from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data import get_sample_id, create_initial_sample
import numpy as np
import tempfile
import os
from fuse.data.ops.op_base import OpBase
from typing import List, Union
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data.datasets.dataset_default import DatasetDefault


class OpFakeLoad(OpBase):
    def __init__(self):
        super().__init__()

    def __call__(self, sample_dict: NDict, **kwargs) -> Union[None, dict, List[dict]]:
        sid = get_sample_id(sample_dict)
        if "case_1" == sid:
            sample_dict.merge(_generate_sample_1())
        elif "case_2" == sid:
            sample_dict.merge(_generate_sample_2())
        elif "case_3" == sid:
            return None
        elif "case_4" == sid:
            sample_1 = create_initial_sample("case_4", "case_4_subcase_1")
            sample_1.merge(_generate_sample_1(41))

            sample_2 = create_initial_sample("case_4", "case_4_subcase_2")
            sample_2.merge(_generate_sample_2(42))

            return [sample_1, sample_2]
        else:
            raise Exception(f"unfamiliar sample_id: {sid}")
        sample_dict = ForMonkeyPatching.identity_transform(sample_dict)
        return sample_dict


class ForMonkeyPatching:
    @staticmethod
    def identity_transform(sample_dict):
        """
        returns the sample as is. The purpose of this is to be monkey-patched in the audit test.
        When it will be modified, the cached samples will become stale,
        as this code is called from within an op, and therefore does not participate in the hash generation.
        """
        return sample_dict


class OpPrintContents(OpBase):
    def __init__(self):
        super().__init__()

    def __call__(self, sample_dict: NDict, **kwargs) -> Union[None, dict, List[dict]]:
        sid = get_sample_id(sample_dict)
        print(f"sid={sid}")
        for k in sample_dict.keypaths():
            print(k)
        print("-------------------------\n")

        return sample_dict


class TestDatasetDefault(unittest.TestCase):
    """
    Test sample caching
    """

    def setUp(self):
        pass

    def test_audit(self):
        tmpdir = tempfile.mkdtemp()
        cache_dirs = [
            os.path.join(tmpdir, "cache_a"),
            os.path.join(tmpdir, "cache_b"),
        ]

        static_pipeline_desc = [
            (OpFakeLoad(), {}),
        ]

        dynamic_pipeline_desc = [
            (OpPrintContents(), {}),
        ]

        static_pl = PipelineDefault(
            "static_pipeline",
            static_pipeline_desc,
        )
        dynamic_pl = PipelineDefault(
            "dynamic_pipeline",
            dynamic_pipeline_desc,
        )

        orig_sample_ids = ["case_1", "case_2"]
        ################ cached + no sample morphing
        cacher = SamplesCacher(
            "dataset_default_audit_test_cache",
            static_pl,
            cache_dirs,
            restart_cache=True,
            audit_rate=1,
            audit_units="samples",
        )

        ds_cached = DatasetDefault(
            orig_sample_ids,
            static_pl,
            dynamic_pipeline=dynamic_pl,
            cacher=cacher,
        )

        ds_cached.create(num_workers=0)
        cached_final_sample_ids = ds_cached.get_all_sample_ids()

        print("a...")
        sample_from_cached = ds_cached[0]
        print("b...")
        sample_from_cached = ds_cached[0]

        def small_change(sample_dict):
            sample_dict["data"]["cc"]["img"][10, 100, 100] += 0.001
            return sample_dict

        ForMonkeyPatching.identity_transform = small_change

        print("c...")
        self.assertRaises(Exception, ds_cached, 0)
        # sample_from_cached = ds_cached[0]

        ForMonkeyPatching.identity_transform = lambda x: x  # return it to previous state

        ########### do it again, and now test the audit_first_sample

        # recreating cacher to change audit parameters
        cacher = SamplesCacher(
            "dataset_default_audit_test_cache",
            static_pl,
            cache_dirs,
            restart_cache=True,
            audit_first_sample=True,
            audit_rate=None,
        )

        ds_cached = DatasetDefault(
            orig_sample_ids,
            static_pl,
            dynamic_pipeline=dynamic_pl,
            cacher=cacher,
        )

        ds_cached.create(num_workers=0)

        ForMonkeyPatching.identity_transform = small_change

        # the first one is expected to raise an exception
        self.assertRaises(Exception, ds_cached, 0)

        ForMonkeyPatching.identity_transform = lambda x: x  # return it to previous state

        ############################## testing audit_first_sample a bit more
        ############################## this time we do the monkey patching only AFTER the first sample was audited (and the staleness will be missed)

        # recreating cacher to change audit params
        cacher = SamplesCacher(
            "dataset_default_audit_test_cache",
            static_pl,
            cache_dirs,
            restart_cache=True,
            audit_first_sample=True,
            audit_rate=None,
        )

        ds_cached = DatasetDefault(
            orig_sample_ids,
            static_pl,
            dynamic_pipeline=dynamic_pl,
            cacher=cacher,
        )

        ds_cached.create(num_workers=0)

        # there is no problem yet, should work well
        sample_from_cached = ds_cached[0]

        # we now monkey patch it, creating a mismatch between the hash and the static pipeline logic
        ForMonkeyPatching.identity_transform = small_change
        # it won't be caught as it didn't happen in the first sample, and we've set audit_rate to None
        sample_from_cached = ds_cached[0]
        sample_from_cached = ds_cached[0]

        banana = 123

    def tearDown(self):
        pass


def _generate_sample_1(seed=1337):
    Seed.set_seed(seed)
    sample = dict(
        data=dict(
            cc=dict(
                img=np.random.rand(30, 200, 200),
                seg=np.random.randint(0, 16, size=(30, 200, 200)),
                dicom_tags=[10, 13, 40, "banana"],
            ),
            mlo=dict(
                img=np.random.rand(30, 200, 200),
                seg=np.random.randint(0, 16, size=(40, 100, 164)),
                dicom_tags=[100, 130, 400, "banana123"],
            ),
            gt_labels_style_1=[1, 3, 100, 12],
            gt_labels_style_2=np.array([3, 4, 10, 12]),
            clinical_info_input=np.random.rand(1000),
        )
    )
    return sample


def _generate_sample_2(seed=1234):
    Seed.set_seed(seed)
    sample = dict(
        data=dict(
            cc=dict(
                img=np.random.rand(10, 100, 100),
                seg=np.random.randint(0, 16, size=(10, 100, 10)),
                dicom_tags=[20, 23, 60, "bananaphone"],
            ),
            mlo=dict(
                img=np.random.rand(30, 200, 200),
                seg=np.random.randint(0, 16, size=(40, 100, 164)),
                dicom_tags=[12, 13, 40, "porcupine123"],
            ),
            gt_labels_style_1=[5, 2, 13, 16],
            gt_labels_style_2=np.array([8, 14, 11, 1]),
            clinical_info_input=np.random.rand(90),
        )
    )
    return sample


if __name__ == "__main__":
    unittest.main()



from typing import Optional, Sequence, Tuple, Union, Hashable
from fuse.eval.metrics.libs.thresholding import Thresholding
from fuse.eval.metrics.utils import PerSampleData

import numpy as np

from .metrics_classification_common import MetricMultiClassDefault


class MetricApplyThresholds(MetricMultiClassDefault):
    def __init__(
        self,
        pred: str,
        class_names: Optional[Sequence[str]] = None,
        operation_point: Union[float, Sequence[Tuple[int, float]], str, None] = None,
        **kwargs
    ):
        """
        :param pred: key name for the model prediction scores
        :param class_names: class names. required for multi-class classifiers
        :param operation_point: Optional. If specified will be used to convert probabilities to class prediction.
                                          Options:
                                          * None - argmax(probability) - default mode
                                          * float - used as a threshold for the positive class in binary classification
                                          * [(<class_index_first>, <threshold_first>), ...,(<class_index_last>, <threshold_last>)] -
                                            the class prediction will be the first class in which the probability cross the threshold.
                                            If no class probability crossed the threshold the predicated class will be set to -1
        """
        super().__init__(
            pred=pred,
            target=None,
            extract_ids=True,
            metric_func=self._apply_thresholds,
            class_names=class_names,
            operation_point=operation_point,
            **kwargs
        )

    def _apply_thresholds(
        self,
        pred: Sequence[np.ndarray],
        ids: Sequence[Hashable],
        operation_point: Union[float, Sequence[Tuple[int, float]], str, None] = None,
    ):

        pred_thresholded = Thresholding.apply_thresholds(pred=pred, operation_point=operation_point)
        # make sure to return the per-sample metric result for the relevant sample ids:
        per_sample_data = PerSampleData(data=pred_thresholded, ids=ids)

        return {"cls_pred": per_sample_data}



from typing import List, Optional, Tuple, Union

from PIL import Image

import numpy as np
import torch
import torchvision.transforms.functional as TTF
import torchvision.transforms as transforms
import skimage
import skimage.transform

from fuse.utils.ndict import NDict

from fuse.data import OpBase


class OpAugAffine2D(OpBase):
    """
    2D affine transformation
    """

    def __init__(self, verify_arguments: bool = True):
        """
        :param verify_arguments: this op expects torch tensor with either 2 or 3 dimensions. Set to False to disable verification
        """
        super().__init__()
        self._verify_arguments = verify_arguments

    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        rotate: float = 0.0,
        translate: Tuple[float, float] = (0.0, 0.0),
        scale: Tuple[float, float] = 1.0,
        flip: Tuple[bool, bool] = (False, False),
        shear: float = 0.0,
        channels: Optional[List[int]] = None,
        interpolation: int = transforms.InterpolationMode.BILINEAR,
    ) -> Union[None, dict, List[dict]]:
        """
        :param key: key to a tensor stored in sample_dict: 2D tensor representing an image to augment, shape [num_channels, height, width] or [height, width]
        :param rotate: angle [-360.0 - 360.0]
        :param translate: translation per spatial axis (number of pixels). The sign used as the direction.
        :param scale: scale factor
        :param flip: flip per spatial axis flip[0] for vertical flip and flip[1] for horizontal flip
        :param shear: shear factor
        :param channels: apply the augmentation on the specified channels. Set to None to apply to all channels.
        :param interpolation: used PIL interpolation method see TTF.affine for details.
        :return: the augmented image
        """
        aug_input = sample_dict[key]

        # verify
        if self._verify_arguments:
            assert isinstance(
                aug_input, torch.Tensor
            ), f"Error: OpAugAffine2D expects torch Tensor, got {type(aug_input)}"
            assert len(aug_input.shape) in [
                2,
                3,
            ], f"Error: OpAugAffine2D expects tensor with 2 or 3 dimensions. got {aug_input.shape}"

        # Support for 2D inputs - implicit single channel
        if len(aug_input.shape) == 2:
            aug_input = aug_input.unsqueeze(dim=0)
            remember_to_squeeze = True
        else:
            remember_to_squeeze = False

        # convert to PIL (required by affine augmentation function)
        if channels is None:
            channels = list(range(aug_input.shape[0]))
        aug_tensor = aug_input
        for channel in channels:
            aug_channel_tensor = aug_input[channel].numpy()
            aug_channel_tensor = Image.fromarray(aug_channel_tensor)
            aug_channel_tensor = TTF.affine(
                aug_channel_tensor,
                angle=rotate,
                scale=scale,
                interpolation=interpolation,
                translate=translate,
                shear=shear,
            )
            if flip[0]:
                aug_channel_tensor = TTF.vflip(aug_channel_tensor)
            if flip[1]:
                aug_channel_tensor = TTF.hflip(aug_channel_tensor)

            # convert back to torch tensor
            aug_channel_tensor = np.array(aug_channel_tensor)
            aug_channel_tensor = torch.from_numpy(aug_channel_tensor)

            # set the augmented channel
            aug_tensor[channel] = aug_channel_tensor

        # squeeze back to 2-dim if needed
        if remember_to_squeeze:
            aug_tensor = aug_tensor.squeeze(dim=0)

        sample_dict[key] = aug_tensor
        return sample_dict


class OpAugCropAndResize2D(OpBase):
    """
    Alternative to rescaling in OpAugAffine2D: center crop and resize back to the original dimensions. if scale is bigger than 1.0. the image first padded.
    """

    def __init__(self, verify_arguments: bool = True):
        """
        :param verify_arguments: this ops expects torch tensor with either 2 or 3 dimensions. Set to False to disable verification
        """
        super().__init__()
        self._verify_arguments = verify_arguments

    def __call__(
        self, sample_dict: NDict, key: str, scale: Tuple[float, float], channels: Optional[List[int]] = None
    ) -> Union[None, dict, List[dict]]:
        """
        :param key: key to a tensor stored in sample_dict: 2D tensor representing an image to augment, shape [num_channels, height, width] or [height, width]
        :param scale: tuple of positive floats
        :param channels: apply augmentation on the specified channels or None for all of them
        :return: the augmented tensor
        """
        aug_input = sample_dict[key]

        # verify
        if self._verify_arguments:
            assert isinstance(
                aug_input, torch.Tensor
            ), f"Error: OpAugCropAndResize2D expects torch Tensor, got {type(aug_input)}"
            assert len(aug_input.shape) in [
                2,
                3,
            ], f"Error: OpAugCropAndResize2D expects tensor with 2 or 3 dimensions. got {aug_input.shape}"

        if len(aug_input.shape) == 2:
            aug_input = aug_input.unsqueeze(dim=0)
            remember_to_squeeze = True
        else:
            remember_to_squeeze = False

        if channels is None:
            channels = list(range(aug_input.shape[0]))
        aug_tensor = aug_input
        for channel in channels:
            aug_channel_tensor = aug_input[channel]

            if scale[0] != 1.0 or scale[1] != 1.0:
                cropped_shape = (
                    int(aug_channel_tensor.shape[0] * scale[0]),
                    int(aug_channel_tensor.shape[1] * scale[1]),
                )
                padding = [[0, 0], [0, 0]]
                for dim in range(2):
                    if scale[dim] > 1.0:
                        padding[dim][0] = (cropped_shape[dim] - aug_channel_tensor.shape[dim]) // 2
                        padding[dim][1] = (cropped_shape[dim] - aug_channel_tensor.shape[dim]) - padding[dim][0]
                aug_channel_tensor_pad = TTF.pad(
                    aug_channel_tensor.unsqueeze(0), (padding[1][0], padding[0][0], padding[1][1], padding[0][1])
                )
                aug_channel_tensor_cropped = TTF.center_crop(aug_channel_tensor_pad, cropped_shape)
                aug_channel_tensor = TTF.resize(aug_channel_tensor_cropped, aug_channel_tensor.shape).squeeze(0)
                # set the augmented channel
                aug_tensor[channel] = aug_channel_tensor

        # squeeze back to 2-dim if needed
        if remember_to_squeeze:
            aug_tensor = aug_tensor.squeeze(dim=0)

        sample_dict[key] = aug_tensor
        return sample_dict


class OpAugSqueeze3Dto2D(OpBase):
    """
    Squeeze selected axis of volume image into channel dimension, in order to fit the 2D augmentation functions
    """

    def __init__(self, verify_arguments: bool = True):
        """
        :param verify_arguments: this ops expects torch tensor with 4 dimensions. Set to False to disable verification
        """
        super().__init__()
        self._verify_arguments = verify_arguments

    def __call__(self, sample_dict: NDict, key: str, axis_squeeze: int) -> NDict:
        """
        :param key: key to a tensor stored in sample_dict: 3D tensor representing an image to augment, shape [num_channels, spatial axis 1, spatial axis 2, spatial axis 3]
        :param axis_squeeze: the axis (1, 2 or 3) to squeeze into channel dimension - typically z axis
        """
        aug_input = sample_dict[key]

        # verify
        if self._verify_arguments:
            assert isinstance(
                aug_input, torch.Tensor
            ), f"Error: OpAugSqueeze3Dto2D expects torch Tensor, got {type(aug_input)}"
            assert (
                len(aug_input.shape) == 4
            ), f"Error: OpAugSqueeze3Dto2D expects tensor with 4 dimensions. got {aug_input.shape}"

        # aug_input shape is [channels, axis_1, axis_2, axis_3]
        if axis_squeeze == 1:
            pass
        elif axis_squeeze == 2:
            aug_input = aug_input.permute((0, 2, 1, 3))
            # aug_input shape is [channels, axis_2, axis_1, axis_3]
        elif axis_squeeze == 3:
            aug_input = aug_input.permute((0, 3, 1, 2))
            # aug_input shape is [channels, axis_3, axis_1, axis_2]
        else:
            raise Exception(f"Error: axis squeeze must be 1, 2, or 3, got {axis_squeeze}")

        aug_output = aug_input.reshape((aug_input.shape[0] * aug_input.shape[1],) + aug_input.shape[2:])

        sample_dict[key] = aug_output
        return sample_dict


class OpAugUnsqueeze3DFrom2D(OpBase):
    """
    Unsqueeze selected axis of volume image from channel dimension, restore the original shape squeezed by OpAugSqueeze3Dto2D
    """

    def __init__(self, verify_arguments: bool = True):
        """
        :param verify_arguments: this ops expects torch tensor with 2 dimensions. Set to False to disable verification
        """
        super().__init__()
        self._verify_arguments = verify_arguments

    def __call__(self, sample_dict: NDict, key: str, axis_squeeze: int, channels: int) -> NDict:
        """
        :param key: key to a tensor stored in sample_dict and squeezed by OpAugSqueeze3Dto2D
        :param axis_squeeze: axis squeeze as specified in OpAugSqueeze3Dto2D
        :param channels: number of channels in the original tensor (before OpAugSqueeze3Dto2D)
        """
        aug_input = sample_dict[key]

        # verify
        if self._verify_arguments:
            assert isinstance(
                aug_input, torch.Tensor
            ), f"Error: OpAugUnsqueeze3DFrom2D expects torch Tensor, got {type(aug_input)}"
            assert (
                len(aug_input.shape) == 3
            ), f"Error: OpAugUnsqueeze3DFrom2D expects tensor with 3 dimensions. got {aug_input.shape}"

        aug_output = aug_input.reshape((channels, aug_input.shape[0] // channels) + aug_input.shape[1:])

        if axis_squeeze == 1:
            pass
        elif axis_squeeze == 2:
            # aug_output shape is [channels, axis_2, axis_1, axis_3]
            aug_output = aug_output.permute((0, 2, 1, 3))
            # aug_input shape is [channels, axis 1, axis 2, axis 3]
        elif axis_squeeze == 3:
            # aug_output shape is [channels, axis_3, axis_1, axis_2]
            aug_output = aug_output.permute((0, 2, 3, 1))
            # aug_input shape is [channels, axis 1, axis 2, axis 3]
        else:
            raise Exception(f"Error: axis squeeze must be 1, 2, or 3, got {axis_squeeze}")

        sample_dict[key] = aug_output
        return sample_dict


class OpCrop3D(OpBase):
    """
    crop to certain size. if the image is smaller than the size then its padded.
    """

    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        output_shape: Tuple[int, int, int],
        z_move=0.5,
        x_move=0.5,
        y_move=0.5,
        fill: int = 0,
    ):
        """
        :param key: key to a tensor stored in sample_dict and get cropped by OpRandomCrop3D
        :param out_size: shape of the output tensor
        :param fill: if the image needs padding then it will be filled with that value
        :param z_move: float between 0 and 1 which is the proportion of the movement that can be done (for x and y is the same idea).
        when they are all set to 0.5 it means we are cropping from the center of the image.
        """
        aug_input = sample_dict[key]
        assert len(aug_input.shape) == len(output_shape)
        assert z_move >= 0 and z_move <= 1
        assert x_move >= 0 and x_move <= 1
        assert y_move >= 0 and y_move <= 1
        depth, height, width = aug_input.shape  # input is in the form [D,H,W]

        aug_tensor = torch.full(output_shape, fill, dtype=torch.float32)

        if depth > output_shape[0]:
            crop_start = round(z_move * (depth - output_shape[0]))
            aug_input = aug_input[crop_start : crop_start + output_shape[0], :, :]
        if height > output_shape[1]:
            crop_start = round(y_move * (height - output_shape[1]))
            aug_input = aug_input[:, crop_start : crop_start + output_shape[1], :]
        if width > output_shape[2]:
            crop_start = round(x_move * (width - output_shape[2]))
            aug_input = aug_input[:, :, crop_start : crop_start + output_shape[2]]

        aug_tensor[:depth, :height, :width] = aug_input
        sample_dict[key] = aug_tensor

        return sample_dict


class OpResizeTo(OpBase):
    """
    Resizes an image into the given dimensions. Currently supports only ndarray
    """

    def __init__(self, channels_first: bool):
        """
        :param channels_first: assign True iff the input is in CxHxW format.
        """
        super().__init__()
        self._channels_first = channels_first

    def __call__(self, sample_dict: NDict, output_shape: Tuple[int], key: str, **kwargs) -> NDict:
        """
        :param key: key to a numpy array or tensor stored in the sample_dict in a H x W x C format.
        :param kwargs: additional arguments to pass to the resize function

        Stores the resized image in sample_dict[key]
        """
        aug_input = sample_dict[key]
        dim = len(aug_input.shape)

        if self._channels_first:
            # Permutes CxHxW -> HxWxC (for skimage's resize)
            perm = self.get_permutation(dim=dim, channels_first=True)
            aug_input = np.transpose(aug_input, axes=perm)

        # Apply Resize
        aug_output = skimage.transform.resize(image=aug_input, output_shape=output_shape, **kwargs)

        if self._channels_first:
            # Permutes back HxWxC -> CxHxW
            perm = self.get_permutation(dim=dim, channels_first=False)
            aug_output = np.transpose(aug_output, axes=perm)

        sample_dict[key] = aug_output

        return sample_dict

    def get_permutation(self, dim: int, channels_first: bool):
        """
        :param dim: tensor's dimension
        :param channels_first: True iff the wanted permutation is: HxWxC -> CxHxW

        Returns the right permutation to:
        channels_first is True -> converting from CxHxW to HxWxC
            i.e: (dim-2, dim-1, 0, 1, ..., dim-3)
        channels_first is False -> converting from HxWxC to CxHxW
            i.e: (2, 3, ..., dim-2, dim-1, 0, 1)
        """

        if dim < 2:
            raise Exception(f"Error, dim ({dim}) must be greater or equal to 2.")

        if channels_first:
            channels = [0]
            hw = [i for i in range(1, dim)]
            perm = hw + channels

        else:
            channels = [dim - 1]
            hw = [i for i in range(0, dim - 1)]
            perm = channels + hw

        perm = tuple(perm)
        return perm


class OpRotation3D(OpBase):
    def __call__(self, sample_dict: NDict, key: str, z_rot: float = 0.0, y_rot: float = 0.0, x_rot: float = 0):
        """
        rotates an input tensor around an axis, when for example z_rot is chosen,
        the rotation is in the x-y plane.
        Note: rotation angles are in relation to the original axis (not the rotated one)
        rotation angles should be given in degrees
        :param aug_input:image input should be in shape [z, y, x] [depth, height, width]
        :param z_rot: angle to rotate x-y plane clockwise
        :param y_rot: angle to rotate x-z plane clockwise
        :param x_rot: angle to rotate z-y plane clockwise
        :return:
        """
        aug_input = sample_dict[key]
        assert len(aug_input.shape) == 3  # will only work for 3d
        if z_rot != 0:
            aug_input = TTF.rotate(aug_input, angle=z_rot)
        if x_rot != 0:
            aug_input = aug_input.permute(dims=(2, 0, 1))
            aug_input = TTF.rotate(aug_input, angle=x_rot)
            aug_input = aug_input.permute(dims=(1, 2, 0))
        if y_rot != 0:
            aug_input = aug_input.permute(dims=(1, 0, 2))
            aug_input = TTF.rotate(aug_input, angle=y_rot)
            aug_input = aug_input.permute(dims=(1, 0, 2))

        sample_dict[key] = aug_input
        return sample_dict



from typing import Dict, Hashable
from fuse.utils.ndict import NDict

"""
helper utilities for creating empty samples, and setting and getting sample_id within samples

A sample is a NDict, which is a special "flavor" of a dictionry, allowing accessing elements within it using x['a.b.c.d'] instead of x['a']['b']['c']['d'],
which is very useful as it allows defining a nested element, or a nested sub-dict using a single string.

The bare minimum that a sample is required to contain are:

'initial_sample_id' - this is an arbitrary (Hashable) identifier. Usually a string, but doesn't have to be.
    It represnts the initial sample_id that was provided before a pipeline was used to process the sample, and potentially use "sample morphing".
    "sample morphing" means that a sample might change during the pipeline execution.
    1. Discard - one type of morphing is that a sample is being discarded. Example use case is discarding an MRI volume because it has too little segmentation info that interests a certain research design.
    2. Split - another type of morphing is that a sample can be split into multiple samples.
    For example, the initial_sample_id represents an entire CT volume, which results in multiple samples, each having the same initial_sample_id, but a different sample_id,
        each representing a slice within the CT volume which contains enough segmentation information

'sample_id' - the sample id, uniquely identifying it. It must be Hashable. Again, usually a string, but doesn't have to be.

"""


def create_initial_sample(initial_sample_id: Hashable, sample_id=None):
    """
    creates an empty sample dict and sets both sample_id and initial_sample_id
    :param sample_id:
    :param initial_sample_id: optional. If not provided, sample_id will be used for it as well
    """
    ans = NDict()

    if sample_id is None:
        sample_id = initial_sample_id

    set_initial_sample_id(ans, initial_sample_id)
    set_sample_id(ans, sample_id)

    return ans


##### sample_id


def get_sample_id_key() -> str:
    """
    return sample id key
    """
    return "data.sample_id"


def get_sample_id(sample: Dict) -> Hashable:
    """
    extracts sample_id from the sample dict
    """
    if get_sample_id_key() not in sample:
        raise Exception
    return sample[get_sample_id_key()]


def set_sample_id(sample: Dict, sample_id: Hashable):
    """
    sets sample_id in an existing sample dict
    """
    sample[get_sample_id_key()] = sample_id


#### dealing with initial sample id - this is related to morphing, and describes the original provided sample_id, prior to the morphing effect


def get_initial_sample_id_key() -> str:
    """
    return initial sample id key
    """
    return "data.initial_sample_id"


def set_initial_sample_id(sample: Dict, initial_sample_id: Hashable):
    """
    sets initial_sample_id in an existing sample dict
    """
    sample[get_initial_sample_id_key()] = initial_sample_id


def get_initial_sample_id(sample: Dict) -> Hashable:
    """
    extracts initial_sample_id from the sample dict
    """
    if get_initial_sample_id_key() not in sample:
        raise Exception
    return sample[get_initial_sample_id_key()]


####


def get_specific_sample_from_potentially_morphed(sample, sample_id):
    if isinstance(sample, dict):
        assert get_sample_id(sample) == sample_id
        return sample
    elif isinstance(sample, list):
        for curr_sample in sample:
            if get_sample_id(curr_sample) == sample_id:
                return curr_sample
        raise Exception(f"Could not find requested sample_id={sample_id}")
    else:
        raise Exception(
            "Expected the sample to be either a dict or a list of dicts. None does not make sense in this context."
        )

    assert False  # should never reach here



from .common import ClassifierFCN, ClassifierFCN3D, ClassifierMLP
from .head_1D_classifier import Head1DClassifier
from .heads_3D import Head3DClassifier
from .head_dense_segmentation import HeadDenseSegmentation
from .head_global_pooling_classifier import HeadGlobalPoolingClassifier



from typing import Dict, Optional, Sequence, Tuple, Union, Hashable
from fuse.eval.metrics.libs.classification import MetricsLibClass
import numpy as np
from sklearn.utils import resample


class Thresholding:
    """
    Methods for thresholding
    """

    @staticmethod
    def apply_thresholds(
        pred: Sequence[np.ndarray], operation_point: Union[float, Sequence[float], None] = None
    ) -> Dict:
        """
        :param pred: sequence of numpy arrays / floats of shape [NUM_CLASSES]
        :param operation_point: Optional. If specified will be used to convert probabilities to class prediction.
                                          Options:
                                          * None - argmax(probability) - default mode
                                          * float - used as a threshold for the positive class in binary classification
                                          * [(<class_index_first>, <threshold_first>), ...,(<class_index_last>, <threshold_last>)] -
                                            the class prediction will be the first class in which the probability cross the threshold.
                                            If no class probability crossed the threshold the predicated class will be set to -1
        """

        return MetricsLibClass.convert_probabilities_to_class(pred, operation_point)



import torch


class LossWarmUp(torch.nn.Module):
    """
    Zero the loss until a defined amount of iterations have passed.

    This is useful for example when you have multiple losses and you want one loss to stabilize before the other is used
    """

    def __init__(self, loss: torch.nn.Module, nof_iterations: int):
        super().__init__()
        self._loss = loss
        self._nof_iterrations = nof_iterations
        self._count = 0

    def forward(self, *args, **kwargs):
        if self._count < self._nof_iterrations:
            self._count += 1
            return torch.tensor(0.0)
        else:
            return self._loss.forward(*args, **kwargs)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import shutil
import tempfile
import unittest
import os
from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess

from fuse.utils.rand.seed import Seed
import fuse.utils.gpu as GPU

if "STOIC21_DATA_PATH" in os.environ:
    from fuse_examples.imaging.classification.stoic21.runner_stoic21 import (
        PATHS,
        TRAIN_COMMON_PARAMS,
        run_train,
        run_infer,
        run_eval,
        INFER_COMMON_PARAMS,
        EVAL_COMMON_PARAMS,
    )


def run_stoic21(root: str) -> None:
    model_dir = os.path.join(root, "model_dir")
    paths = {
        "model_dir": model_dir,
        "data_dir": PATHS["data_dir"],
        "cache_dir": os.path.join(root, "cache_dir"),
        "data_split_filename": os.path.join(root, "split.pkl"),
        "inference_dir": os.path.join(model_dir, "infer_dir"),
        "eval_dir": os.path.join(model_dir, "eval_dir"),
    }

    train_common_params = TRAIN_COMMON_PARAMS
    train_common_params["trainer.num_epochs"] = 2
    infer_common_params = INFER_COMMON_PARAMS

    analyze_common_params = EVAL_COMMON_PARAMS

    GPU.choose_and_enable_multiple_gpus(1)

    Seed.set_seed(0, False)  # previous test (in the pipeline) changed the deterministic behavior to True
    run_train(paths, train_common_params)
    run_infer(paths, infer_common_params)
    results = run_eval(paths, analyze_common_params)

    assert "metrics.auc" in results


@unittest.skipIf(
    "STOIC21_DATA_PATH" not in os.environ, "define environment variable 'STOIC21_DATA_PATH' to run this test"
)
class ClassificationStoic21TestCase(unittest.TestCase):
    def setUp(self):
        self.root = tempfile.mkdtemp()

    def test_template(self):
        run_in_subprocess(run_stoic21, self.root, timeout=1800)

    def tearDown(self):
        # Delete temporary directories
        shutil.rmtree(self.root)


if __name__ == "__main__":
    unittest.main()



# version
from fuse.version import __version__

from fuse.utils.ndict import NDict
from fuse.utils.data.collate import CollateToBatchList, uncollate

from fuse.utils.rand.param_sampler import Uniform, RandInt, RandBool, Choice, draw_samples_recursively
from fuse.utils.rand.seed import Seed

set_seed = Seed.set_seed
from fuse.utils.file_io.file_io import read_dataframe, save_dataframe
from fuse.utils.data.collate import CollateToBatchList, uncollate



import unittest

from typing import Optional, Union, List
from fuse.utils.ndict import NDict

from fuse.data.ops.op_base import OpBase, OpReversibleBase, op_call, op_reverse
from fuse.data import create_initial_sample
from fuse.data import OpRepeat
from fuse.data.ops.ops_aug_common import OpRandApply, OpSample, OpSampleAndRepeat, OpRepeatAndSample
from fuse.utils.rand.param_sampler import Choice, RandBool, RandInt, Uniform
from fuse.utils import Seed


class OpArgsForTest(OpReversibleBase):
    def __init__(self):
        super().__init__()

    def __call__(self, sample_dict: NDict, op_id: Optional[str], **kwargs) -> Union[None, dict, List[dict]]:
        return {"op_id": op_id, "kwargs": kwargs}

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        return {"op_id": op_id}


class OpBasicSetter(OpBase):
    """
    A basic op for testing, which sets sample_dict[key] to set_key_to_val
    """

    def __init__(self):
        super().__init__()

    def __call__(self, sample_dict: NDict, key, set_key_to_val, **kwargs) -> Union[None, dict, List[dict]]:
        sample_dict[key] = set_key_to_val
        return sample_dict


class TestOpsAugCommon(unittest.TestCase):
    def test_op_sample(self):
        Seed.set_seed(0)
        a = {
            "a": 5,
            "b": [3, RandInt(1, 5), 9],
            "c": {"d": 3, "f": [1, 2, RandBool(0.5), {"h": RandInt(10, 15)}]},
            "e": {"g": Choice([6, 7, 8])},
        }
        op = OpSample(OpArgsForTest())
        result = op_call(op, {}, "op_id", **a)
        b = result["kwargs"]
        call_op_id = result["op_id"]
        # make sure the same op_id passed to internal op
        self.assertEqual(call_op_id, "op_id")

        # make srgs sampled correctly
        self.assertEqual(a["a"], a["a"])
        self.assertEqual(b["b"][0], a["b"][0])
        self.assertEqual(b["b"][2], a["b"][2])
        self.assertEqual(b["c"]["d"], a["c"]["d"])
        self.assertEqual(b["c"]["f"][1], a["c"]["f"][1])
        self.assertIn(b["b"][1], [1, 2, 3, 4, 5])
        self.assertIn(b["c"]["f"][2], [True, False])
        self.assertIn(b["c"]["f"][3]["h"], [10, 11, 12, 13, 14, 15])
        self.assertIn(b["e"]["g"], [6, 7, 8])

        # make sure the same op_id passed also in reverse
        result = op_reverse(op, {}, "", "", "op_id")
        reversed_op_id = result["op_id"]
        self.assertEqual(reversed_op_id, "op_id")

    def test_op_sample_and_repeat(self):
        Seed.set_seed(1337)
        sample_1 = create_initial_sample(0)
        op = OpSampleAndRepeat(OpBasicSetter(), [dict(key="data.input.img"), dict(key="data.gt.seg")])
        sample_1 = op_call(op, sample_1, op_id="testing_sample_and_repeat", set_key_to_val=Uniform(3.0, 6.0))

        Seed.set_seed(1337)
        sample_2 = create_initial_sample(0)
        op = OpSample(OpRepeat(OpBasicSetter(), [dict(key="data.input.img"), dict(key="data.gt.seg")]))
        sample_2 = op_call(op, sample_2, op_id="testing_sample_and_repeat", set_key_to_val=Uniform(3.0, 6.0))

        self.assertEqual(sample_1["data.input.img"], sample_1["data.gt.seg"])
        self.assertEqual(sample_1["data.input.img"], sample_2["data.input.img"])

    def test_op_repeat_and_sample(self):
        Seed.set_seed(1337)
        sample_1 = create_initial_sample(0)
        op = OpRepeatAndSample(OpBasicSetter(), [dict(key="data.input.img"), dict(key="data.gt.seg")])
        sample_1 = op_call(op, sample_1, op_id="testing_sample_and_repeat", set_key_to_val=Uniform(3.0, 6.0))

        Seed.set_seed(1337)
        sample_2 = create_initial_sample(0)
        op = OpRepeat(
            OpSample(
                OpBasicSetter(),
            ),
            [dict(key="data.input.img"), dict(key="data.gt.seg")],
        )
        sample_2 = op_call(op, sample_2, op_id="testing_sample_and_repeat", set_key_to_val=Uniform(3.0, 6.0))

        self.assertEqual(sample_1["data.input.img"], sample_2["data.input.img"])
        self.assertEqual(sample_1["data.gt.seg"], sample_2["data.gt.seg"])

    def test_op_rand_apply(self):
        """
        Test OpRandApply
        """
        Seed.set_seed(0)
        op = OpRandApply(OpArgsForTest(), 0.5)

        def sample(op):
            return "kwargs" in op_call(op, {}, "op_id", a=5)

        # test range
        self.assertIn(sample(op), [True, False])

        # test generate more than a single number
        Seed.set_seed(0)
        values = [sample(op) for _ in range(4)]
        self.assertIn(True, values)
        self.assertIn(False, values)

        # test probs
        Seed.set_seed(0)
        op = OpRandApply(OpArgsForTest(), 0.99)
        count = 0
        for _ in range(1000):
            if sample(op) == True:
                count += 1
        self.assertGreaterEqual(count, 980)


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

======================================

Collection of useful functions to implement FuseMedML pytorch lightning based module and train loop
"""
import traceback
from typing import Any, Dict, List, OrderedDict, Sequence, Union
from statistics import mean
from fuse.data.utils.sample import get_sample_id_key
from fuse.utils.data.collate import uncollate
import pandas as pd

import torch
import pytorch_lightning as pl
from pytorch_lightning.callbacks import ModelCheckpoint
from .pl_epoch_summary import ModelEpochSummary

from fuse.utils import NDict
from fuse.dl.losses.loss_base import LossBase
from fuse.eval import MetricBase
from fuse.eval.metrics.utils import PerSampleData


def model_checkpoint_callbacks(model_dir: str, best_epoch_source: Union[Dict, List[Dict]]) -> List[pl.Callback]:
    """
    Create list of pl.callbacks that saves checkpoints using (pl.callbacks.ModelCheckpoint) and print per epoch summary (fuse.dl.lightning.pl_epoch_summary.ModelEpochSummary).
    :param model_dir: path to save checkpoints and summary
    :param best_epoch_source: either a dict with arguments to pass to ModelCheckpoint or list dicts to for multiple ModelCheckpoint callbacks (to monitor and save checkpoints for more then one metric).
    """
    callbacks = []
    # checkpoints
    if not isinstance(best_epoch_source, list):
        best_epoch_source = [best_epoch_source]
    for checkpoint_to_monitor in best_epoch_source:
        if "dirpath" not in checkpoint_to_monitor:
            checkpoint_to_monitor["dirpath"] = model_dir
        if "filename" not in checkpoint_to_monitor:
            checkpoint_to_monitor["filename"] = "best_epoch"
            if len(best_epoch_source) > 1:
                checkpoint_to_monitor["auto_insert_metric_name"] = True

        model_checkpoint = ModelCheckpoint(**checkpoint_to_monitor)
        model_checkpoint_display = ModelEpochSummary(
            dirpath=checkpoint_to_monitor["dirpath"],
            monitor=checkpoint_to_monitor.get("monitor", None),
            mode=checkpoint_to_monitor.get("mode", "min"),
        )
        callbacks.append(model_checkpoint)
        callbacks.append(model_checkpoint_display)

    # last epoch checkpoint
    callbacks.append(ModelCheckpoint(dirpath=model_dir, filename="last_epoch", save_last=True))
    return callbacks


def convert_predictions_to_dataframe(predictions: List[NDict]) -> pd.DataFrame:
    """list of batch_dict to a dataframe"""
    assert len(predictions) > 0

    values = {}
    predictions_per_sample = []
    for elem in predictions:
        predictions_per_sample += uncollate(elem)
    keys = predictions_per_sample[0].keypaths()
    for key in keys:
        values[key] = [elem[key] for elem in predictions_per_sample]

    df = pd.DataFrame(values)
    return df


def step_losses(losses: Dict[str, LossBase], batch_dict: NDict) -> torch.Tensor:
    """
    Compute losses per step (batch) in pl.LightningModule.<training/validation/test>_step()
    :param losses: dict of FuseMedML style losses
    :param batch_dict: FuseMedML batch_dict including data and model outputs
    :return: total_loss (sum all losses results). The values for tracking purpose will be stored in batch_dict['losses']
    """
    total_loss = None
    for loss_name, loss_function in losses.items():
        current_loss_result = loss_function(batch_dict)
        batch_dict["losses." + loss_name] = current_loss_result.data.item()
        # sum all losses for backward
        if total_loss is None:
            total_loss = current_loss_result
        else:
            total_loss += current_loss_result

    if total_loss is not None:
        batch_dict["losses.total_loss"] = total_loss.data.item()

    return total_loss


def step_metrics(metrics: OrderedDict[str, MetricBase], batch_dict: NDict) -> None:
    """
    Collect data to compute per epoch metrics
    :param metrics: dictionary of metrics
    :param batch_dict: FuseMedML batch_dict including data, losses and model outputs
    :return: None
    """
    for _, metric in metrics.items():
        # handle batch doesn't return a value, the actual value of the metric is per epoch
        metric.collect(batch_dict)


def step_extract_predictions(prediction_keys: Sequence[str], batch_dict: NDict) -> Dict[str, Any]:
    """
    Extract the specified perdictions from batch_dict (torch Tensors will be detached, moved to cpu and coverted to numpy)
    :param prediction_keys: the keys to extract
    :param batch_dict: FuseMedML batch_dict including data and model outputs
    """
    outputs = {}
    sample_ids = batch_dict[get_sample_id_key()]
    if isinstance(sample_ids, torch.Tensor):
        sample_ids = list(sample_ids.detach().cpu().numpy())
    outputs["id"] = sample_ids
    for key in prediction_keys:
        output = batch_dict[key]
        if isinstance(output, torch.Tensor):
            output = output.detach().cpu().numpy()
        outputs[key] = output

    return outputs


def epoch_end_compute_and_log_losses(pl: pl.LightningModule, mode: str, batch_losses: Sequence[Dict]) -> None:
    """
    On epoch end average out the batch losses and log the averaged losses
    :param pl: LightiningModule. Used for logging.
    :param mode: prefix to add to each loss name (when logging), typically validation/train/test
    :param batch_losses: list of batch_dict["losses"] as added by 'epoch_losses'
    :return: None
    """
    keys = batch_losses[0].keys()
    for key in keys:
        losses = []
        for elem in batch_losses:
            if isinstance(elem[key], torch.Tensor):
                losses.extend(elem[key].detach().cpu().tolist())
            else:
                losses.append(elem[key])
        loss = mean(losses)
        pl.log(f"{mode}.losses.{key}", loss, on_epoch=True)


def epoch_end_compute_and_log_metrics(pl: pl.LightningModule, mode: str, metrics: OrderedDict[str, MetricBase]) -> None:
    """
    On epoch end compute and log per epoch metrics
    :param pl: LightiningModule. Used for logging.
    :param mode: prefix to add to each metric name (when logging), typically validation/train/test
    :param metrics: Dict of FuseMedML style metircs
    :return: None
    """
    # compute metrics
    epoch_results = NDict()
    # compute metrics and keep the results
    for metric_name, metric in metrics.items():
        try:
            metric_result = metric.eval(epoch_results)
        except:
            track = traceback.format_exc()
            print(f"Metric {metric_name} process() func failed. Setting results to None")
            print(track)
            metric_result = None

        epoch_results[f"metrics.{metric_name}"] = metric_result
        metric.reset()

    # log metrics
    for key in epoch_results.keypaths():
        if epoch_results[key] is not None and not isinstance(epoch_results[key], (PerSampleData)):
            pl.log(f"{mode}.{key}", epoch_results[key], on_epoch=True)



from typing import List
import numpy as np
from fuse.utils.ndict import NDict
import torch

# TODO: support custom _object_requires_hdf5_single
#      maybe even more flexible (knowing key name etc., patterns, explicit name, regular expr.)

# TODO: should we require OrderedDict?? and for the internal dicts as well ??
# TODO: maybe it's better to flatten the dictionaries first


def _object_requires_hdf5_recurse(curr: NDict, str_base="") -> List[str]:
    """
    Iterates on keys and checks
    """
    keys = curr.keypaths()
    ans = []
    for k in keys:
        data = curr[k]
        if _object_requires_hdf5_single(data):
            ans.append(k)
    return ans


# def _PREV__object_requires_hdf5_recurse(curr: NDict, str_base='') -> List[str]:
#     """
#     Recurses (only into dicts!) and returns a list of keys that require storing into HDF5
#     (which allows reading only sub-parts)

#     :return: a list of keys as strings, e.g. ['data.cc.img', 'data.mlo.img']
#     """
#     #print('str_base=', str_base)
#     if _object_requires_hdf5_single(curr):
#         return str_base

#     if isinstance(curr, dict):
#         ans = []
#         for k,d in curr.items():
#             curr_ans = _object_requires_hdf5_recurse(
#                 d, str_base+'.'+k if str_base!='' else k,
#             )
#             if curr_ans is None:
#                 pass
#             elif isinstance(curr_ans, list):
#                 ans.extend(curr_ans)
#             else:
#                 ans.append(curr_ans)
#         return ans

#     return None


def _object_requires_hdf5_single(obj, minimal_ndarray_size=100):
    ans = isinstance(obj, np.ndarray) and (obj.size > minimal_ndarray_size)

    if isinstance(obj, torch.Tensor):
        raise Exception(
            "You need to cast to tensor in the dynamic pipeline as it takes a lot of time pickling torch.Tensor"
        )

    # if ans:
    #    print(f'found hfd5 requiring object! shape={obj.shape}, size={obj.size}')
    return ans



import os
from fuse.utils.config_tools import get_config_function


def load_config(config: dict) -> dict:
    _curr_dir = os.path.dirname(os.path.abspath(__file__))
    _base_conf = os.path.join(_curr_dir, "base_conf_example.py")
    config = get_config_function(_base_conf)(config)
    config["banana"] = 123
    return config



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Union, Tuple

import numpy as np
import skimage
import torch
import cv2


def match_img_to_input(im: np.ndarray, input: np.ndarray):
    """
    Resize an im to the input shape
    :param im: 2D image, can be numpy or Tensor
    :param input: input image, shape [channel, W, H]. numpy or Tensor
    :return: resized im
    """
    if isinstance(im, torch.Tensor):
        im = im.detach().cpu().numpy()

    if len(im.shape) < 3:
        im = np.expand_dims(im, axis=0)

    if (input.shape[1], input.shape[2]) != (im.shape[1], im.shape[2]):
        resize_im = np.zeros((im.shape[0], input.shape[1], input.shape[2]))
        for c in range(im.shape[0]):
            resize_im[c, :] = cv2.resize(im[c, :], (input.shape[2], input.shape[1]), interpolation=cv2.INTER_NEAREST)
        im = resize_im

    return im


def pad_ndimage(
    ndimage: np.ndarray, outer_height: int, outer_width: int, pad_value: Union[float, int]
) -> Tuple[np.ndarray, int, int]:
    """
    Pastes input ndimage in the middle of a larger one
    :param ndimage:         2-dim ndimage
    :param outer_height:    final outer height
    :param outer_width:     final outer width
    :param pad_value:       value for padding around inner image
    :return:
    """
    inner_height, inner_width = ndimage.shape
    h_offset = int((outer_height - inner_height) / 2.0)
    w_offset = int((outer_width - inner_width) / 2.0)
    outer_image = np.ones((outer_height, outer_width), dtype=ndimage.dtype) * pad_value
    outer_image[h_offset : h_offset + inner_height, w_offset : w_offset + inner_width] = ndimage
    return outer_image, h_offset, w_offset


def block_reduce_resize(img: np.ndarray, target_shape: Tuple[int, int] = (10, 5), func=np.max) -> np.ndarray:
    """
    Reduces an image by applying 'func' param on blocks, to yield a target shape.
    :param img:             2D ndarray, shape [height, width]
    :param target_shape:    target shape [height, width]
    :param func:            callable, e.g., np.max or np.average
    :return:  reduced ndimage
    """
    block_size = (int(img.shape[0] / target_shape[0]), int(img.shape[1] / target_shape[1]))
    reduced_img = skimage.measure.block_reduce(img, block_size=block_size, func=func)

    # Block reduce produces a near final result, but its dimensions might not match target_shape exactly. So resize!
    resized_reduced_img = skimage.transform.resize(
        reduced_img,
        target_shape,
        mode="edge",
        anti_aliasing=False,
        anti_aliasing_sigma=None,
        preserve_range=True,
        order=0,
    )
    return resized_reduced_img


def preserve_range_resize(img: np.ndarray, target_shape: Tuple[int, int] = (10, 5)) -> np.ndarray:
    """
    Resizes a 2D ndarray without anti-aliasing and while preserving dtype and range
    :param img:             2D ndarray
    :param target_shape:    target shape
    :return:
    """
    resized_img = skimage.transform.resize(
        img, target_shape, mode="edge", anti_aliasing=False, anti_aliasing_sigma=None, preserve_range=True, order=0
    )
    return resized_img


def align_ecc(
    img1: np.ndarray,
    img2: np.ndarray,
    num_iterations: int = 400,
    termination_eps: float = 1e-4,
    transformation: int = None,
) -> np.ndarray:
    """
    Aligns two images using ECC estimation.
    Returns transformed img2 to match img1.
    :param img1:      ndarray of shape [height, width] and dtype=float32
    :param img2:      ndarray of shape [height, width] and dtype=float32
    :param num_iterations:  number of iterations until stopping criteria
    :param termination_eps: epsilon stopping criteria
    :param transformation: type of transformation to perform. If None (default), cv2.MOTION_AFFINE is performed.
    :return   transformed img2

    The implementation was moved to a separate class; see AlignMapECC. This function serves for backward
    compatibility
    """

    try:
        import cv2
        from fuse.utils.imaging.align.utils_align_ecc import AlignMapECC

        transformation = transformation or cv2.MOTION_AFFINE

        aligner = AlignMapECC(
            transformation_type=transformation, num_iterations=num_iterations, termination_eps=termination_eps
        )

        aligner.align(img1, img2)

        return aligner.img2_aligned
    except Exception as e:
        import logging

        logging.getLogger("Fuse").error("Cannot align images", e)
        return img2


def resize_image(image: np.ndarray, resize_to: Tuple[int, int]):
    """
    Resizes image to resize_to dimensions
    :param image: ndarray of shape [height, width]
    :param resize_to: (expected_height, expected_width)
    :return: image resized to resize_to
    """

    import cv2

    image_original_h = image.shape[0]
    image_original_w = image.shape[1]

    # Resize
    if image.shape[0] > resize_to[0]:
        h_ratio = resize_to[0] / image_original_h
    else:
        h_ratio = 1
    if image.shape[1] > resize_to[1]:
        w_ratio = resize_to[1] / image_original_w
    else:
        w_ratio = 1

    resize_ratio = min(h_ratio, w_ratio)
    if resize_ratio != 1:
        return cv2.resize(image, dsize=(int(image.shape[1] * resize_ratio), int(image.shape[0] * resize_ratio)))

    return image



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest


import random

from fuse.utils import Uniform, Choice, RandInt, RandBool, draw_samples_recursively, Seed


class TestParamSampler(unittest.TestCase):
    """
    Test ParamSampleBase sub classes
    """

    def test_uniform(self):
        Seed.set_seed(0)
        min = random.random() * 1000
        max = random.random() * 1000 + min
        uniform = Uniform(min, max)
        value = uniform.sample()

        # test range
        self.assertGreaterEqual(value, min)
        self.assertLessEqual(uniform.sample(), max)

        # test generate more than a single number
        self.assertNotEqual(value, uniform.sample())

        # test fixed per seed
        Seed.set_seed(1234)
        value0 = uniform.sample()
        Seed.set_seed(1234)
        value1 = uniform.sample()
        self.assertEqual(value0, value1)

    def test_randint(self):
        Seed.set_seed(0)
        min = random.randint(0, 1000)
        max = random.randint(0, 1000) + min
        randint = RandInt(min, max)
        value = randint.sample()

        # test range
        self.assertGreaterEqual(value, min)
        self.assertLessEqual(randint.sample(), max)

        # test generate more than a single number
        self.assertNotEqual(value, randint.sample())

        # test fixed per seed
        Seed.set_seed(1234)
        value0 = randint.sample()
        Seed.set_seed(1234)
        value1 = randint.sample()
        self.assertEqual(value0, value1)

    def test_randbool(self):
        Seed.set_seed(0)
        randbool = RandBool(0.5)
        value = randbool.sample()

        # test range
        self.assertIn(value, [True, False])

        # test generate more than a single number
        Seed.set_seed(0)
        values = [randbool.sample() for _ in range(4)]
        self.assertIn(True, values)
        self.assertIn(False, values)

        # test fixed per seed
        Seed.set_seed(1234)
        value0 = randbool.sample()
        Seed.set_seed(1234)
        value1 = randbool.sample()
        self.assertEqual(value0, value1)

        # test probs
        Seed.set_seed(0)
        randbool = RandBool(0.99)
        count = 0
        for _ in range(1000):
            if randbool.sample() == True:
                count += 1
        self.assertGreaterEqual(count, 980)

    def test_choice(self):
        Seed.set_seed(0)
        lst = list(range(1000))
        choice = Choice(lst)
        value = choice.sample()

        # test range
        self.assertIn(value, lst)

        # test generate more than a single number
        self.assertNotEqual(value, choice.sample())

        # test fixed per seed
        Seed.set_seed(1234)
        value0 = choice.sample()
        Seed.set_seed(1234)
        value1 = choice.sample()
        self.assertEqual(value0, value1)

        # test probs
        Seed.set_seed(0)
        probs = [0.01 / 999] * 1000
        probs[5] = 0.99
        choice = Choice(lst, probs)
        count = 0
        for _ in range(1000):
            if choice.sample() == 5:
                count += 1
        self.assertGreaterEqual(count, 980)

    def test_draw_samples_recursively(self):
        Seed.set_seed(0)
        a = {
            "a": 5,
            "b": [3, RandInt(1, 5), 9],
            "c": {"d": 3, "f": [1, 2, RandBool(0.5), {"h": RandInt(10, 15)}]},
            "e": {"g": Choice([6, 7, 8])},
        }
        b = draw_samples_recursively(a)

        self.assertEqual(a["a"], a["a"])
        self.assertEqual(b["b"][0], a["b"][0])
        self.assertEqual(b["b"][2], a["b"][2])
        self.assertEqual(b["c"]["d"], a["c"]["d"])
        self.assertEqual(b["c"]["f"][1], a["c"]["f"][1])
        self.assertIn(b["b"][1], [1, 2, 3, 4, 5])
        self.assertIn(b["c"]["f"][2], [True, False])
        self.assertIn(b["c"]["f"][3]["h"], [10, 11, 12, 13, 14, 15])
        self.assertIn(b["e"]["g"], [6, 7, 8])


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
import unittest

from typing import List
from fuse.data.ops.op_base import op_call, op_reverse
from fuse.utils.ndict import NDict
import pandas as pd
import torch
import numpy as np
from fuse.data.ops.ops_cast import OpToNumpy, OpToTensor


class TestOpsCast(unittest.TestCase):
    def test_op_to_tensor(self):
        """
        Test OpToTensor __call__ and reverse
        """
        op = OpToTensor()
        sample = NDict(
            {
                "sample_id": 7,
                "values": {
                    "val_np": np.array([7, 8, 9]),
                    "val_torch": torch.tensor([1, 2, 3]),
                    "val_int": 3,
                    "val_float": 3.5,
                    "str": "hi!",
                },
            }
        )

        sample = op_call(op, sample, "_.test_id", key="values.val_np")
        self.assertIsInstance(sample["values.val_np"], torch.Tensor)
        self.assertTrue((sample["values.val_np"] == torch.tensor([7, 8, 9])).all())
        self.assertIsInstance(sample["values.val_int"], int)

        sample = op_call(op, sample, "_.test_id", key=["values.val_torch", "values.val_float"])
        self.assertIsInstance(sample["values.val_torch"], torch.Tensor)
        self.assertIsInstance(sample["values.val_float"], torch.Tensor)
        self.assertTrue((sample["values.val_torch"] == torch.tensor([1, 2, 3])).all())
        self.assertEqual(sample["values.val_float"], torch.tensor(3.5))
        self.assertIsInstance(sample["values.val_int"], int)

        sample = op_reverse(
            op, sample, key_to_follow="values.val_np", key_to_reverse="values.val_np", op_id="_.test_id"
        )
        self.assertIsInstance(sample["values.val_np"], np.ndarray)

    def test_op_to_numpy(self):
        """
        Test OpToNumpy __call__ and reverse
        """
        op = OpToNumpy()
        sample = NDict(
            {
                "sample_id": 7,
                "values": {
                    "val_np": np.array([7, 8, 9]),
                    "val_torch": torch.tensor([1, 2, 3]),
                    "val_int": 3,
                    "val_float": 3.5,
                    "str": "hi!",
                },
            }
        )

        sample = op_call(op, sample, "_.test_id", key="values.val_torch")
        self.assertIsInstance(sample["values.val_torch"], np.ndarray)
        self.assertTrue((sample["values.val_torch"] == np.array([1, 2, 3])).all())
        self.assertIsInstance(sample["values.val_int"], int)

        sample = op_call(op, sample, "_.test_id", key=["values.val_np", "values.val_float"])
        self.assertIsInstance(sample["values.val_np"], np.ndarray)
        self.assertIsInstance(sample["values.val_float"], np.ndarray)
        self.assertTrue((sample["values.val_np"] == np.array([7, 8, 9])).all())
        self.assertEqual(sample["values.val_float"], np.array(3.5))
        self.assertIsInstance(sample["values.val_int"], int)

        sample = op_reverse(
            op, sample, key_to_follow="values.val_torch", key_to_reverse="values.val_torch", op_id="_.test_id"
        )
        self.assertIsInstance(sample["values.val_torch"], torch.Tensor)


if __name__ == "__main__":
    unittest.main()



import cv2

from fuse.data.ops.op_base import OpBase
from fuseimg.utils.typing.key_types_imaging import DataTypeImaging
from fuseimg.data.ops.ops_common_imaging import OpApplyTypesImaging
from fuse.utils.ndict import NDict

# import SimpleITK as sitk


def no_op(input_tensor):
    return input_tensor


def draw_grid_3d_op(
    input_tensor, start_slice=0, end_slice=None, line_color=255, thickness=10, type_=cv2.LINE_4, pxstep=50
):
    """
    Draws a grid pattern.
    #todo: it is possible to change this function to support both 2d and 3d

    :param input_tensor: a numpy array, either HW  format for grayscale or HWC
    if HWC and C >4 then assumed to be a 3d grayscale

    :param line_color:
    :param thickness:
    :param type_:
    :param pxstep:
    :return:
    """

    # grid = sitk.GridSource(outputPixelType=sitk.sitkUInt16, size=input_tensor.shape, sigma=(0.5, 0.5,0.5), gridSpacing=(100.0, 100.0, 100.0), gridOffset=(0.0, 0.0, 0.0), spacing=(0.2, 0.2, 0.2))
    # grid = sitk.GetArrayFromImage(grid)

    if end_slice is None:
        end_slice = input_tensor.shape[2] - 1

    for s in range(start_slice, end_slice + 1):
        x = pxstep
        y = pxstep
        while x < input_tensor.shape[1]:
            cv2.line(
                input_tensor[..., s],
                (x, 0),
                (x, input_tensor.shape[0]),
                color=line_color,
                lineType=type_,
                thickness=thickness,
            )
            x += pxstep

        while y < input_tensor.shape[0]:
            cv2.line(
                input_tensor[..., s],
                (0, y),
                (input_tensor.shape[1], y),
                color=line_color,
                lineType=type_,
                thickness=thickness,
            )
            y += pxstep

    return input_tensor


# Define function to draw a grid
def draw_grid(im, grid_size):
    # Draw grid lines
    # im = Image.fromarray(im)
    # im = im.astype(np.float32)
    for i in range(0, im.shape[1], grid_size):
        cv2.line(im, (i, 0), (i, im.shape[0]), color=(255,))
    for j in range(0, im.shape[0], grid_size):
        cv2.line(im, (0, j), (im.shape[1], j), color=(255,))
    return im


class OpDrawGrid(OpBase):
    """
    draws a 2d grid on the input tensor for debugging
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, sample_dict: NDict, key: str, grid_size):
        img = sample_dict[key]
        draw_grid(img, grid_size=grid_size)

        sample_dict[key] = img
        return sample_dict


op_draw_grid_img = OpApplyTypesImaging({DataTypeImaging.IMAGE: (OpDrawGrid(), {})})



from typing import Optional, Tuple, List
import numpy as np
from torch import Tensor
import skimage
import torch
import torchvision.transforms.functional as TTF

from fuse.utils.ndict import NDict

from fuse.data.ops.op_base import OpBase
from skimage import measure
from fuseimg.utils.typing.key_types_imaging import DataTypeImaging
from fuseimg.data.ops.ops_common_imaging import OpApplyTypesImaging


def sanity_check_HWC(input_tensor):
    if 3 != input_tensor.ndim:
        raise Exception(f"expected 3 dim tensor, instead got {input_tensor.shape}")
    assert input_tensor.shape[2] < input_tensor.shape[0]
    assert input_tensor.shape[2] < input_tensor.shape[1]


def sanity_check_CHW(input_tensor):
    if 3 != input_tensor.ndim:
        raise Exception(f"expected 3 dim tensor, instead got {input_tensor.shape}")
    assert input_tensor.shape[0] < input_tensor.shape[1]
    assert input_tensor.shape[0] < input_tensor.shape[2]


class OpHWCToCHW(OpBase):
    """
    HWC (height, width, channel) to CHW (channel, height, width)
    """

    def __call__(self, sample_dict: NDict, key: str) -> NDict:
        """
        :param key: key to torch tensor of shape [H, W, C]
        """
        input_tensor: Tensor = sample_dict[key]

        sanity_check_HWC(input_tensor)
        input_tensor = input_tensor.permute(dims=(2, 0, 1))
        sanity_check_CHW(input_tensor)

        sample_dict[key] = input_tensor
        return sample_dict


class OpCHWToHWC(OpBase):
    """
    CHW (channel, height, width) to HWC (height, width, channel)
    """

    def __call__(self, sample_dict: NDict, key: str) -> NDict:
        """
        :param key: key to torch tensor of shape [C, H, W]
        """
        input_tensor: Tensor = sample_dict[key]

        sanity_check_CHW(input_tensor)
        input_tensor = input_tensor.permute(dims=(1, 2, 0))
        sanity_check_HWC(input_tensor)

        sample_dict[key] = input_tensor
        return sample_dict


class OpSelectSlice(OpBase):
    """
    select one slice from the input tensor,
    from the first dimension of a >2 dimensional input
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, sample_dict: NDict, op_id: Optional[str], key: str, slice_idx: int):
        """
        :param slice_idx: the index of the selected slice from the 1st dimmention of an input tensor
        """

        img = sample_dict[key]
        if len(img.shape) < 3:
            return sample_dict

        img = img[slice_idx]
        sample_dict[key] = img
        return sample_dict


class OpResizeAndPad2D(OpBase):
    """
    Resize and Pad a 2D image
    """

    def __init__(self, number_of_channels: int = 1, pad_value: int = 0, **kwargs):
        super().__init__(**kwargs)
        self.number_of_channels = number_of_channels
        self.pad_value = pad_value

    def pad_image(
        self, inner_image: np.ndarray, padding: Tuple[float, float], resize_to: Tuple[int, int]
    ) -> np.ndarray:
        """
        pads image to requested size ,
        pads both side equally by the same input padding size (left = right = padding[1] , up = down= padding[0] )  ,
        padding default value is zero or minimum value in normalized target range
        :param inner_image: image of shape [H, W, C] of type numpy float32
        :param padding: required padding [x,y]
        :param resize_to: original requested resolution
        :param normalized_target_range: requested normalized image pixels range
        :param number_of_channels: number of color channels in the image
        :return: padded image
        """
        image = self.pad_inner_image(
            inner_image,
            outer_height=resize_to[0] + 2 * padding[0],
            outer_width=resize_to[1] + 2 * padding[1],
            pad_value=self.pad_value,
        )
        return image

    def pad_inner_image(self, image: np.ndarray, outer_height: int, outer_width: int, pad_value: float) -> np.ndarray:
        """
        Pastes input image in the middle of a larger one
        :param image:        image of shape [H, W, C]
        :param outer_height: final outer height
        :param outer_width:  final outer width
        :param pad_value:    value for padding around inner image
        :number_of_channels  final number of channels in the image
        :return:             padded image
        """
        inner_height, inner_width = image.shape[0], image.shape[1]
        h_offset = int((outer_height - inner_height) / 2.0)
        w_offset = int((outer_width - inner_width) / 2.0)
        if self.number_of_channels > 1:
            outer_image = np.ones((outer_height, outer_width, self.number_of_channels), dtype=image.dtype) * pad_value
            outer_image[h_offset : h_offset + inner_height, w_offset : w_offset + inner_width, :] = image
        elif self.number_of_channels == 1:
            outer_image = np.ones((outer_height, outer_width), dtype=image.dtype) * pad_value
            outer_image[h_offset : h_offset + inner_height, w_offset : w_offset + inner_width] = image
        return outer_image

    def resize_image(self, inner_image: np.ndarray, resize_to: Tuple[int, int]) -> np.ndarray:
        """
        resize image to the required resolution
        :param inner_image: image of shape [H, W, C]
        :param resize_to: required resolution [height, width]
        :return: resized image
        """
        inner_image_height, inner_image_width = inner_image.shape[0], inner_image.shape[1]
        if inner_image_height > resize_to[0]:
            h_ratio = resize_to[0] / inner_image_height
        else:
            h_ratio = 1
        if inner_image_width > resize_to[1]:
            w_ratio = resize_to[1] / inner_image_width
        else:
            w_ratio = 1

        resize_ratio = min(h_ratio, w_ratio)
        if resize_ratio != 1:
            inner_image = skimage.transform.resize(
                inner_image,
                output_shape=(int(inner_image_height * resize_ratio), int(inner_image_width * resize_ratio)),
                mode="reflect",
                anti_aliasing=True,
            )
        return inner_image

    def __call__(self, sample_dict: NDict, key: str, resize_to: Tuple, padding: Tuple):
        """
        :param resize_to:               new size of input images, keeping proportions
        :param padding:                 required padding size [x,y]
        """

        img = sample_dict[key]
        # resize
        if resize_to is not None:
            img = self.resize_image(img, resize_to)

        # padding
        if padding is not None:
            img = self.pad_image(img, padding, resize_to)
        sample_dict[key] = img
        return sample_dict


class OpFindBiggestNonEmptyBbox2D(OpBase):
    """
    Finds the the biggest connected component bounding box in the image that is non empty (dark)
    """

    def __init__(self, dark_area_threshold: int = 10, blocks_num: int = 30, **kwargs):
        super().__init__(**kwargs)
        self.dark_area_threshold = dark_area_threshold
        self.blocks_num = blocks_num

    def find_biggest_non_emtpy_bbox(self, img: np.ndarray) -> Tuple[int, int, int, int]:
        """
        split the images into blocks, each block containing (1/30 x 1/30) of the image.
        All blocks above a threshold (10) are considered non-empty.
        Then, the biggest connected component (at the blocks level) is extracted, and its axis-aligned bbox is returned.
        :param img: Image instance , expected 2d integer grayscale image where 0 is black background color
        :param dark_area_threshold: defines grayscale level from which lower is consider dark / outside of body
        :param blocks_num: number of blocks the algorithm split the images into
        :return: four coordinates of the bounding box
        """
        bl_rows = img.shape[0] // self.blocks_num
        bl_cols = img.shape[1] // self.blocks_num

        rows_starts = list(range(self.blocks_num))
        cols_starts = list(range(self.blocks_num))

        rows_starts = list(map(lambda x: x * bl_rows, rows_starts))
        cols_starts = list(map(lambda x: x * bl_cols, cols_starts))

        cells = np.zeros((self.blocks_num, self.blocks_num))

        for ri, r in enumerate(rows_starts):
            for ci, c in enumerate(cols_starts):
                r_end = min(img.shape[0] - 1, r + bl_rows)
                c_end = min(img.shape[1] - 1, c + bl_cols)
                cells[ri, ci] = np.mean(img[r:r_end, c:c_end])

        cells_binary = np.zeros((self.blocks_num, self.blocks_num), dtype=bool)
        cells_binary[cells > self.dark_area_threshold] = True
        blobs_labels = measure.label(cells_binary, background=0)
        regions = measure.regionprops(blobs_labels)

        regions_areas = [r.area for r in regions]
        if len(regions) < 1:
            print("Warning: could not crop properly! fallbacking to full image")
            return 0, 0, img.shape[0] - 1, img.shape[1] - 1

        for i, r in enumerate(regions):
            if r.label == 0:
                regions_areas[i] = -1
                break
        max_ind = np.argmax(regions_areas)
        bbox = regions[max_ind].bbox

        full_img_bbox = [bbox[0] * bl_rows, bbox[1] * bl_cols, (bbox[2] + 1) * bl_rows, (bbox[3] + 1) * bl_cols]
        minr, minc, maxr, maxc = full_img_bbox
        maxr = min(maxr, img.shape[0] - 1)
        maxc = min(maxc, img.shape[1] - 1)

        return minr, minc, maxr, maxc

    def __call__(self, sample_dict: NDict, key: str):
        """ """

        img = sample_dict[key]
        aabb = self.find_biggest_non_emtpy_bbox(img)
        img = img[aabb[0] : aabb[2], aabb[1] : aabb[3]].copy()
        sample_dict[key] = img
        return sample_dict


class OpFlipBrightSideOnLeft2D(OpBase):
    """
    Returns an image where the brigheter half side is on the left, flips the image if the condition does nt hold.
    """

    def __init__(self, max_pixel_value: float = 255.0, dark_region_ratio: float = 15.0, **kwargs):
        super().__init__(**kwargs)
        self.max_pixel_value = max_pixel_value
        self.dark_region_ratio = dark_region_ratio

    def check_bright_side_is_left(self, image: np.ndarray) -> bool:
        """
        checks if the bright side is in the left of the image
        :param image: numpy image , expected 2d grayscale image
        :param max_pixel_value: maximum possible value in the image grayscale format
        :param dark_region_ratio: the raito of possible grayscale values which are considered dark
        :return: True iff the bright side is left
        """
        cols = image.shape[1]
        left_side = image[:, : cols // 2]
        right_side = image[:, cols // 2 :]
        dark_region = self.max_pixel_value / self.dark_region_ratio
        return np.count_nonzero(left_side < dark_region) < np.count_nonzero(right_side < dark_region)

    def __call__(self, sample_dict: NDict, key: str):
        """
        :param image: numpy image , expected 2d grayscale image
        :return: image where the breast is in the left
        """

        image = sample_dict[key]
        if not self.check_bright_side_is_left(image):  # orig
            image = np.fliplr(image)
            sample_dict[key] = image
        return sample_dict


op_select_slice_img_and_seg = OpApplyTypesImaging(
    {DataTypeImaging.IMAGE: (OpSelectSlice(), {}), DataTypeImaging.SEG: (OpSelectSlice(), {})}
)


class OpPad(OpBase):
    """
    Pad the given image on all the sides. Supports Tensor & ndarray.
    """

    def __call__(
        self, sample_dict: NDict, key: str, padding: List[int], fill: int = 0, mode: str = "constant", **kwargs
    ):
        """
        Pad values
        :param key: key to an image in sample_dict - either torch tensor or ndarray
        :param padding: padding on each border. can be differerniate each border by passing a list.
        :param fill: if mode = 'constant', pads with fill's value.
        :param padding_mode: see torch's & numpy's pad functions for more details.
        :param kwargs: numpy's pad function give supports to more arguments. See it's docs for more details.
        """

        img = sample_dict[key]

        if torch.is_tensor(img):
            processed_img = TTF.pad(img, padding, fill, mode)

        elif isinstance(img, np.ndarray):
            # kwargs['constant_values'] = fill
            processed_img = np.pad(img, pad_width=padding, mode=mode, constant_values=fill, **kwargs)

        else:
            raise Exception(f"Error: OpPad expects Tensor or nd.array object, but got {type(img)}.")

        sample_dict[key] = processed_img
        return sample_dict



from typing import Any, Dict, Optional, Sequence, Tuple, Union

import numpy as np

from scipy.spatial import distance
from scipy.spatial.distance import directed_hausdorff

import matplotlib.pyplot as plt


class MetricsSegmentation:
    @staticmethod
    def get_tf_ft_values_from_bool_array_with_weights(u: np.ndarray, v: np.ndarray, w: Optional[Dict[int, np.ndarray]]):
        """
        Calculates the false positive and true negative between two ndarray , while having different weight for each pixel
        :param u: single sample prediction matrix ( np.ndarray of any shape) per sample
        :param v: target mask ( np.ndarray of any shape ) per sample
        :param w: Optional dictionary - key = label ,value = weight per pixel . Each element is  float in range [0-1]
        """
        not_u = 1.0 - u
        not_v = 1.0 - v
        not_u = w * not_u
        u = w * u
        nft = (not_u * v).sum()
        ntf = (u * not_v).sum()
        return (nft, ntf)

    @staticmethod
    def dice(pred: np.ndarray, target: np.ndarray, pixel_weight: Optional[Dict[int, np.ndarray]] = None) -> Dict:
        """
        Compute dice similarity score (2*|X&Y| / (|X|+|Y|)) using sklearn , pred and target should be of same shape
        Supports multiclass (semantic segmentation)
        :param pred: single sample prediction matrix ( np.ndarray of any shape of type int/bool ) per sample
        :param target: target mask ( np.ndarray of any shape of type int/bool) per sample
        :param pixel_weight: Optional dictionary - key = label ,value = weight per pixel . Each element is  float in range [0-1]
        :return dice score
        """

        labels = np.unique(target)
        labels = labels[labels != 0]
        scores = {}
        for label in labels:
            mask_pred = pred == label
            mask_gt = target == label
            label = str(int(label))
            gt_empty = np.sum(mask_gt) == 0
            pred_empty = np.sum(mask_pred) == 0

            # dice not defined if both are empty ( 0/0 situation)
            if gt_empty and pred_empty:
                scores[label] = 1.0
            else:
                if pixel_weight is None or pixel_weight[label] is None:
                    scores[label] = 1.0 - distance.dice(mask_pred.flatten(), mask_gt.flatten())
                else:
                    scores[label] = 1.0 - distance.dice(
                        mask_pred.flatten(), mask_gt.flatten(), pixel_weight[label].flatten()
                    )

        return scores

    @staticmethod
    def iou_jaccard(pred: np.ndarray, target: np.ndarray, pixel_weight: Optional[Dict[int, np.ndarray]] = None) -> Dict:
        """
        Calculates intersection over union (iou) (|X&Y| / | XUY |) score based on predicted and target segmentation mask
        Supports multiclass (semantic segmentation)
        :param pred: single sample prediction matrix ( np.ndarray of any shape of type int/bool) per sample
        :param target: target mask ( np.ndarray of any shape of type int/bool) per sample
        :param pixel_weight: Optional dictionary - key = label ,value = weight per pixel . Each element is  float in range [0-1]
        :return: iou score
        """
        # extract info for the plot
        labels = np.unique(target)
        labels = labels[labels != 0]
        scores = {}
        for label in labels:
            mask_pred = pred == label
            mask_gt = target == label
            label = str(int(label))
            gt_empty = np.sum(mask_gt) == 0
            pred_empty = np.sum(mask_pred) == 0

            # iou jaccard not defined if both are empty ( 0/0 situation)
            if gt_empty and pred_empty:
                scores[label] = 1.0
            else:
                if pixel_weight is None or pixel_weight[label] is None:
                    intersection = np.logical_and(pred, target)
                    union = np.logical_or(pred, target)
                    iou_score = np.sum(intersection) / np.sum(union)
                    scores[label] = iou_score
                else:
                    ntt = (pred * target * pixel_weight[label]).sum()
                    (nft, ntf) = MetricsSegmentation.get_tf_ft_values_from_bool_array_with_weights(
                        pred, target, pixel_weight[label]
                    )
                    iou_score = ntt / (nft + ntf + ntt)
                    scores[label] = iou_score
        return scores

    @staticmethod
    def overlap(pred: np.ndarray, target: np.ndarray, pixel_weight: Optional[Dict[int, np.ndarray]] = None) -> Dict:
        """
        Calculates overlap score (|X&Y| / min(|X|,|Y|)) based on predicted and target segmentation mask
        Supports multiclass (semantic segmentation)
        :param pred: single sample prediction matrix ( np.ndarray of any shape of type int/bool) per sample
        :param target: target mask ( np.ndarray of any shape of type int/bool) per sample
        :param pixel_weight: Optional dictionary - key = label ,value = weight per pixel . Each element is  float in range [0-1]
        :return: overlap score
        """
        # extract info for the plot
        labels = np.unique(target)
        labels = labels[labels != 0]
        scores = {}
        for label in labels:
            mask_pred = pred == label
            mask_gt = target == label
            label = str(int(label))
            gt_empty = np.sum(mask_gt) == 0
            pred_empty = np.sum(mask_pred) == 0

            # overlap not defined if both are empty ( 0/0 situation)
            if gt_empty and pred_empty:
                scores[label] = 1.0
            elif gt_empty or pred_empty:
                scores[label] = 0.0
            else:
                if pixel_weight is None or pixel_weight[label] is None:
                    intersection = np.logical_and(pred, target)
                    overlap = np.sum(intersection) / min(np.sum(pred), np.sum(target))
                    scores[label] = overlap
                else:
                    intersection = pred * target * pixel_weight[label]
                    overlap = np.sum(intersection) / min(
                        np.sum(pred * pixel_weight[label]), np.sum(target * pixel_weight[label])
                    )
                    scores[label] = overlap
        return scores

    @staticmethod
    def hausdorff_2d_distance(pred: np.ndarray, target: np.ndarray) -> Dict:
        """
        Calculates 2D hausdorff distance based on predicted and target segmentation mask - works for 2D array only!
        Supports multiclass (semantic segmentation)
        :param pred: single sample prediction matrix 2D  array ( np.ndarray  [H, W] of type int/bool) per sample
        :param target: target mask 2D array ( np.ndarray [H, W]  of type int/bool) per sample
        :return: hausdorff distance
        """
        assert len(pred.shape) == 2 or len(target.shape) == 2
        labels = np.unique(target)
        labels = labels[labels != 0]
        scores = {}
        for label in labels:
            mask_pred = pred == label
            mask_gt = target == label
            label = str(int(label))
            mask_pred = mask_pred.astype(int)
            mask_gt = mask_gt.astype(int)
            gt_empty = np.sum(mask_gt) == 0
            pred_empty = np.sum(mask_pred) == 0

            # hausdorff not defined if both are empty ( 0/0 situation)
            if gt_empty and pred_empty:
                scores[label] = 1.0
            else:
                hausdorff1 = directed_hausdorff(mask_pred, mask_gt)[0]
                hausdorff2 = directed_hausdorff(mask_gt, mask_pred)[0]
                hausdorff = max(hausdorff1, hausdorff2)
                scores[label] = hausdorff
        return scores

    @staticmethod
    def pixel_accuracy(
        pred: np.ndarray, target: np.ndarray, pixel_weight: Optional[Dict[int, np.ndarray]] = None
    ) -> Dict:
        """
        Calculates pixel accuracy score (|X&Y| / |Y| ) based on predicted and target segmentation mask
        Supports multiclass (semantic segmentation)
        :param pred: single sample prediction matrix ( np.ndarray of any shape of type int/bool) per sample
        :param target: target mask ( np.ndarray of any shape of type int/bool) per sample
        :param pixel_weight: Optional dictionary - key = label ,value = weight per pixel . Each element is  float in range [0-1]
        :return: pixel accuracy score
        """
        # extract info for the plot
        labels = np.unique(target)
        labels = labels[labels != 0]
        scores = {}
        for label in labels:
            mask_pred = pred == label
            mask_gt = target == label
            label = str(int(label))
            gt_empty = np.sum(mask_gt) == 0
            pred_empty = np.sum(mask_pred) == 0

            # dice not defined if both are empty ( 0/0 situation)
            if gt_empty and pred_empty:
                scores[label] = 1.0
            else:
                if pixel_weight is None or pixel_weight[label] is None:
                    intersection = np.logical_and(pred, target)
                    pixel_accuracy = np.sum(intersection) / np.sum(target)
                    scores[label] = pixel_accuracy
                else:
                    intersection = pred * target * pixel_weight[label]
                    pixel_accuracy = np.sum(intersection) / np.sum(target * pixel_weight[label])
                    scores[label] = pixel_accuracy
        return scores



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest

from tempfile import mkstemp
import pandas as pds


from fuse.utils.file_io.file_io import read_dataframe
from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.utils.export import ExportDataset
from fuse.data.pipelines.pipeline_default import PipelineDefault


class TestDatasetExport(unittest.TestCase):
    def test_export_to_dataframe(self):
        # datainfo
        data = {"sample_id": ["a", "b", "c", "d", "e"], "values": [7, 4, 9, 2, 4], "not_important": [12] * 5}
        df = pds.DataFrame(data)

        # create simple pipeline
        op = OpReadDataframe(df)
        pipeline = PipelineDefault("test", [(op, {})])

        # create dataset
        dataset = DatasetDefault(data["sample_id"], dynamic_pipeline=pipeline)
        dataset.create()

        df = df.set_index("sample_id")

        # export dataset - only get
        export_df = ExportDataset.export_to_dataframe(dataset, ["values"])
        export_df = export_df.set_index("data.sample_id")
        for sid in data["sample_id"]:
            self.assertEqual(export_df.loc[sid]["values"], df.loc[sid]["values"])

        # export dataset - including save
        _, filename = mkstemp(suffix=".gz")
        _ = ExportDataset.export_to_dataframe(dataset, ["values"], output_filename=filename)
        export_df = read_dataframe(filename)
        export_df = export_df.set_index("data.sample_id")
        for sid in data["sample_id"]:
            self.assertEqual(export_df.loc[sid]["values"], df.loc[sid]["values"])


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional, Sequence, Union


from fuse.utils.rand.param_sampler import RandBool, draw_samples_recursively

from fuse.data.ops.op_base import OpBase, OpReversibleBase, op_call, op_reverse
from fuse.data.ops.ops_common import OpRepeat

from fuse.utils.ndict import NDict


class OpRandApply(OpReversibleBase):
    def __init__(self, op: OpBase, probability: float):
        """
        Randomly apply the op (according to the given probability)
        :param op: op
        """
        super().__init__()
        self._op = op
        self._param_sampler = RandBool(probability=probability)

    def __call__(self, sample_dict: NDict, op_id: Optional[str], **kwargs) -> Union[None, dict, List[dict]]:
        """
        See super class
        """
        apply = self._param_sampler.sample()
        sample_dict[op_id] = apply
        if apply:
            sample_dict = op_call(self._op, sample_dict, f"{op_id}.apply", **kwargs)
        return sample_dict

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        """
        See super class
        """
        apply = sample_dict[op_id]
        if apply:
            sample_dict = op_reverse(self._op, sample_dict, key_to_reverse, key_to_follow, f"{op_id}.apply")

        return sample_dict


class OpSample(OpReversibleBase):
    """
    recursively searches for ParamSamplerBase instances in kwargs, and replaces the drawn values inplace before calling to op.__call__()

    For example:
    from fuse.utils.rand.param_sampler import Uniform
    pipeline_desc = [
        #...
        OpSample(OpRotateImage()), {'rotate_angle': Uniform(0.0,360.0)}
        #...
    ]

    OpSample will draw from the Uniform distribution, and will (e.g.) pass rotate_angle=129.43 to OpRotateImage call.

    """

    def __init__(self, op: OpBase):
        """
        :param op: op
        """
        super().__init__()
        self._op = op

    def __call__(self, sample_dict: NDict, op_id: Optional[str], **kwargs) -> Union[None, dict, List[dict]]:
        """
        See super class
        """
        sampled_kwargs = draw_samples_recursively(kwargs)
        return op_call(self._op, sample_dict, op_id, **sampled_kwargs)

    def reverse(self, sample_dict: NDict, key_to_reverse: str, key_to_follow: str, op_id: Optional[str]) -> dict:
        """
        See super class
        """
        return op_reverse(self._op, sample_dict, key_to_reverse, key_to_follow, op_id)


class OpSampleAndRepeat(OpSample):
    """
    First sample kwargs and then repeat op with the exact same sampled arguments.
    This is the equivalent of using OpSample around an OpRepeat.

    Typical usage pattern:
    pipeline_desc = [
        (OpSampleAndRepeat(
            [op to run],
            [a list of dicts describing what to repeat] ),
            [a dictionary describing values that should be the same in all repeated invocations, may include sampling operations like Uniform, RandBool, etc.] ),
    ]

    Example use case:
        randomly choose a rotation angle, and then use the same randomly selected rotation angle
         for both an image and its respective ground truth segmentation map

    from fuse.utils.rand.param_sampler import Uniform
    pipeline_desc = [
        #...
        (OpSampleAndRepeat(OpRotateImage(),
            [dict(key='data.input.img'), dict(key='data.gt.seg')] ),
            dict(angle=Uniform(0.0,360.0)) #this will be drawn only once and the same value will be passed on both OpRotateImage invocation
        ),
        #...
    ]

    #note: this is a convenience op, and it is the equivalent of composing OpSample and OpRepeat yourself.
    The previous example is effectively the same as:

    pipeline_desc = [
        #...
        OpSample(OpRepeat(OpRotateImage(
            [dict(key='data.input.img'), dict(key='data.gt.seg')]),
            dict(angle=Uniform(0.0,360.0)))
        ),
        #...
    ]

    note: see OpRepeatAndSample if you are searching for the opposite flow - drawing a different value per repeat invocation
    """

    def __init__(self, op: OpBase, kwargs_per_step_to_add: Sequence[dict]):
        """
        :param op: the operation to repeat with the same sampled arguments
        :param kwargs_per_step_to_add: sequence of arguments (kwargs format) specific for a single repetition. those arguments will be added/overide the kwargs provided in __call__() function.
        """
        super().__init__(OpRepeat(op, kwargs_per_step_to_add))


class OpRepeatAndSample(OpRepeat):
    """
    Repeats an op multiple times, each time with different kwargs, and draws random values from distribution SEPARATELY per invocation.

    An example usage scenario, let's say that you train a model which is expected get as input two images:
    'data.input.adult_img' which is an image of an adult, and
    'data.input.child_img' which is an image of a child

    the model task is to predict if this child is a child of this adult (a binary classification task).

    The model is expected to work on images that are rotated to any angle, and there's no reason to suspect correlation between the rotation of the two images,
    so you would like to use rotation augmentation separately for the two images.

    In this case you could do:

    pipeline_desc = [
        #...
        (OpRepeatAndSample(OpRotateImage(),
            [dict(key='data.input.adult_img'), dict(key='data.input.child_img')]),
            dict(dict(angle=Uniform(0.0,360.0))   ### this will be drawn separately per OpRotateImage invocation
        )
        #...
    ]


    note: see also OpSampleAndRepeat if you are looking for the opposite flow, drawing the same value and using it for all repeat invocations
    """

    def __init__(self, op: OpBase, kwargs_per_step_to_add: Sequence[dict]):
        """
        :param op: the operation to repeat
        :param kwargs_per_step_to_add: sequence of arguments (kwargs format) specific for a single repetition. those arguments will be added/overide the kwargs provided in __call__() function.
        """
        super().__init__(OpSample(op), kwargs_per_step_to_add)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Dict, Callable

import torch

from fuse.utils.ndict import NDict


class ModelWrapSeqToDict(torch.nn.Module):
    """
    Fuse model wrapper for wrapping torch modules and passing through Fuse
    """

    def __init__(
        self,
        *,  # preventing positional args
        model: torch.nn.Module = None,
        model_inputs: Sequence[str] = None,
        model_outputs: Sequence[str] = None,
        pre_forward_processing_function: Callable = None,
        post_forward_processing_function: Callable = None,
    ) -> None:
        """
        Default Fuse model wrapper - extracts batch_dict data from model_inputs keys and calls forward of model.
        Then, puts forward's output into batch_dict model_output keys.
        For pre processing of the batch dict data use pre_forward_processing_function.
        For post processing of the forward's output data use post_forward_processing_function.

        :param model: The model to wrap
        :param model_inputs: sequence of keys in batch dict to transfer into model.forward function
            for example: model_inputs=('data.input.input_0.tensor',)
        :param model_outputs: keys in batch dict to save the outputs of model.forward()
            for example: model_outputs=('output.output_0',)
        :param pre_forward_processing_function: utility function to process input data before forward is called (after it's extracted from batch_dict)
        :param post_forward_processing_function: utility function to process forward's output data before it is saved into batch_dict



        """
        super().__init__()
        self.model_inputs = model_inputs
        self.model = model
        self.pre_forward_processing_function = pre_forward_processing_function
        self.post_forward_processing_function = post_forward_processing_function
        self.model_outputs = model_outputs

    def forward(self, batch_dict: NDict) -> Dict:
        # convert input to the model's expected input
        model_input = [batch_dict[conv_input] for conv_input in self.model_inputs]

        # convert input to model expected input
        if self.pre_forward_processing_function is not None:
            model_input = self.pre_forward_processing_function(model_input)

        # run the model
        model_output = self.model.forward(*model_input)

        # convert output of model to Fuse expected output
        if self.post_forward_processing_function is not None:
            model_output = self.post_forward_processing_function(model_output)

        if len(self.model_outputs) == 1:
            batch_dict[self.model_outputs[0]] = model_output
        else:
            for i, output_name in enumerate(self.model_outputs):
                batch_dict[output_name] = model_output[i]

        return batch_dict


class ModelWrapDictToSeq(torch.nn.Module):
    """
    Fuse model wrapper for wrapping fuse pytorch model and make him be in basic format- input is tensor and output is tensor
    The user need to provide the input and output keys of the fuse model
    """

    def __init__(self, fuse_model: torch.nn.Module, output_key: str, input_key: str):
        super().__init__()
        self.model = fuse_model
        self.output_key = output_key
        self.input_key = input_key

    def forward(self, input: torch.tensor):
        batch_dict = NDict()
        # find input key
        batch_dict[self.input_key] = input
        # feed fuse model with dict as he excpect
        ans_ndict = self.model(batch_dict)
        # extract model output from dict
        output = ans_ndict[self.output_key]
        return output



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from __future__ import annotations
from _collections_abc import dict_items

import copy
import types
import numpy
import torch
from typing import Any, Callable, Iterator, Optional, Sequence, Union, List


class NDict(dict):
    """N(ested)Dict - wraps a python dict, and allows to access nested elements via '.' separated key desc

    NOTE: assumes that all keys (including nested) are:
        1. strings
        2. do not contain '.' within a single key, as '.' is used as a special symbol for accessing deeper level nested dict.

    For example:

    x = dict(
        a = dict(
            b = 10,
            c = 12,
            d = dict(
                zz = 'abc',
            )
        ),
        c = 100,
    )

    nx = NDict(x)
    nx['a.b'] = 14
    assert nx['a.b'] == 14

    if the result is a non-leaf, you will get a NDict instance, for example
    assert nx['a']['d.zz'] == 'abc'

    In addition to standard python dict methods, implements:
    * flatten
    * to_dict
    * combine

    """

    def __init__(self, dict_like: Union[dict, tuple, types.GeneratorType, NDict, None] = None):
        """
        :param d: the data with which to populate the nested dictionary, in case of NDict it acts as view constructor,
            otherwise we just set all the keys and values using the setitem function
        """

        self._stored = dict()

        if dict_like is None:
            self._stored = {}
        elif isinstance(dict_like, NDict):
            self._stored = dict_like._stored
        else:
            if not isinstance(dict_like, dict):
                dict_like = dict(dict_like)
            for k, d in dict_like.items():
                self[k] = d

    # NDict custom methods
    def to_dict(self) -> dict:
        """
        converts to standard python dict
        :param copy: set to None (default) to get access to the internal stored dict
        """
        return self._stored

    def clone(self, deepcopy: bool = True) -> NDict:
        """
        does a deep or a shallow copy, shallow copy means only top level keys are copied and the values are only referenced
        in deep copy, all values are copied recursively
        :param deepcopy: if true, does deep copy, otherwise does shalow copy
        """
        if not deepcopy:
            return NDict(copy.copy(self._stored))
        else:
            return NDict(copy.deepcopy(self._stored))

    def flatten(self) -> dict:
        """
        flattens the dictionary
        :returns dict

        For example:

        nx = NDict({'a': {'b': 14, 'c': 12}, 'c': 100, 'z': {'foo': {'boo': 111}}})
        print(nx.flatten())
        {'a.b': 14, 'a.c': 12, 'c': 100, 'z.foo.boo': 111}

        #you can use it to get a list of the flat keys:
        print(nx.flatten().keys())
        """

        all_keys = {}
        for key in self._stored:
            if isinstance(self._stored[key], dict):
                all_sub_keys = NDict(self[key]).flatten()
                keys_to_add = {f"{key}.{sub_key}": all_sub_keys[sub_key] for sub_key in all_sub_keys}
                all_keys.update(keys_to_add)
            else:
                all_keys[key] = self._stored[key]

        return all_keys

    def keypaths(self) -> List[str]:
        """
        returns a list of keypaths (i.e. "a.b.c.d") to all values in the nested dict
        """
        return list(self.flatten().keys())

    def keys(self) -> dict_items:
        return self._stored.keys()

    def values(self) -> dict_items:
        return self._stored.values()

    def items(self) -> dict_items:
        return self._stored.items()

    def merge(self, other: dict) -> NDict:
        """
        inplace merge between self and other.
        """
        other_flat = NDict(other).flatten()
        for k, v in other_flat.items():
            self[k] = v

        return

    def __getitem__(self, key: str) -> Any:
        """
        traverses the nested dict by the path extracted from spliting the key on '.', if key not found,
        optionally shows the possible closest options
        :param key: dot delimited keypath into the nested dict
        """
        nested_key = key.split(".")
        if not nested_key[0] in self._stored:
            raise NestedKeyError(key, self)

        value = self._stored
        for sub_key in nested_key:
            if isinstance(value, dict) and sub_key in value:
                value = value.get(sub_key)
            else:
                raise NestedKeyError(key, self)

        return value

    def __setitem__(self, key: str, value: Any) -> None:
        """
        go over the the dictionary according to the path, create the nodes that does not exist
        :param key: the keypath
        :param value: value to set
        """
        # if value is dictionary add to self key by key to avoid from keys with delimeter "."
        if isinstance(value, dict):
            for sub_key in value:
                self[f"{key}.{sub_key}"] = value[sub_key]
            return

        nested_key = key.split(".")
        element = self._stored
        for key in nested_key[:-1]:
            if key not in element:
                element[key] = {}
            element = element[key]

        # set the value
        element[nested_key[-1]] = value

    def __delitem__(self, key: str) -> None:
        nested_key = key.split(".")
        steps = len(nested_key)
        value = self._stored
        for step_idx, sep_key in enumerate(nested_key):
            if step_idx < steps - 1:
                value = value[sep_key]
            else:  # last step
                del value[sep_key]

    def get_closest_key(self, key: str) -> str:
        """
        For a given keypath, returns the longest valid keypath in the current nested dict
        :param key: a full keypath with dot delimiter
        """
        partial_key = []
        partial_ndict = self._stored
        parts = key.split(".")
        for k in parts:
            if isinstance(partial_ndict, dict) and k in partial_ndict:
                partial_key.append(k)
                partial_ndict = partial_ndict[k]
            else:
                break
        return ".".join(partial_key)

    def pop(self, key: str) -> Any:
        """
        return the value nested_dict[key] and remove the key from the dict.
        :param nested_dict: the dictionary
        :param key: the key to return and remove
        """
        res = self[key]
        del self[key]
        return res

    def indices(self, indices: Union[torch.Tensor, numpy.ndarray]) -> dict:
        """
        Extract the specified indices from each element in the dictionary (if possible)
        :param nested_dict: input dict
        :param indices: indices to extract. Either list of indices or boolean numpy array
        :return: NDict with only the required indices
        """
        new_dict = {}
        all_keys = self.keypaths()
        for key in all_keys:
            try:
                value = self[key]
                if isinstance(value, numpy.ndarray) or isinstance(value, torch.Tensor):
                    new_value = value[indices]
                elif isinstance(value, Sequence):
                    new_value = [item for i, item in enumerate(value) if indices[i]]
                else:
                    new_value = value
                new_dict[key] = new_value
            except:
                print(f"failed to process key {key}")
                raise
        return new_dict

    def apply_on_all(self, apply_func: Callable, *args: Any) -> None:
        all_keys = self.keypaths()
        for key in all_keys:
            new_value = apply_func(self[key], *args)
            self[key] = new_value

    def __reduce__(self) -> Union[str, tuple]:
        return super().__reduce__()

    def __iter__(self) -> Iterator:
        return iter(self._stored)

    def __len__(self) -> int:
        return len(self._stored)

    def __str__(self) -> str:
        return str(self._stored)

    def __repr__(self) -> str:
        return repr(self._stored)

    def __contains__(self, o: str) -> bool:
        return o == self.get_closest_key(o)

    def get(self, key: str, default_value: Any = None) -> Any:
        if key not in self:
            return default_value
        return self[key]

    def get_multi(self, keys: Optional[List[str]] = None) -> NDict:
        if keys is None:
            keys = self.keypaths()  # take all keys

        ans = NDict()

        for k in keys:
            curr = self[k]
            ans[k] = curr

        return ans

    def print_tree(self) -> None:
        """
        print the inner structure of the nested dict with a tree-like structure.
        """
        self._print_tree_static(self._stored)

    @staticmethod
    def _print_tree_static(data_dict: dict, level: int = 0) -> None:
        """
        static-method to print the inner structure of a dict in a tree-like structure.
        """
        if type(data_dict) == dict:
            keys = data_dict.keys()
            level += 1
            for key in keys:
                print("---" * level, key)
                NDict._print_tree_static(data_dict[key], level)


class NestedKeyError(KeyError):
    def __init__(self, key: str, d: NDict) -> None:
        partial_key = d.get_closest_key(key)
        if partial_key == "":
            partial_ndict = d
        else:
            partial_ndict = d[partial_key]

        if isinstance(partial_ndict, NDict):
            options = str([f"{partial_key}.{k}" for k in partial_ndict.keypaths()])
            error_str = f"Error: key {key} does not exist\n. Possible keys on the same branch are: {options}. All keys {d.keypaths()}"
        else:
            error_str = f"Error: key {key} does not exist\n. Closest key is: {partial_key}. All keys: {d.keypaths()}"
        print(error_str)
        super().__init__(error_str)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from functools import partial
from glob import glob
import os
from typing import Hashable, Optional, Sequence, Tuple


import numpy as np
import torch
import skimage
import skimage.transform


from fuse.utils import NDict
from fuse.utils.rand.param_sampler import RandBool, RandInt, Uniform

from fuse.data import DatasetDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data import PipelineDefault, OpToTensor
from fuse.data.ops.op_base import OpBase
from fuse.data.ops.ops_aug_common import OpSample, OpRandApply
from fuse.data.ops.ops_common import OpConcat, OpLambda, OpLookup, OpToOneHot
from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.ops.ops_cast import OpToFloat, OpToInt, OpToNumpy
from fuse.data.utils.sample import get_sample_id

from fuseimg.data.ops.aug.geometry import OpAugAffine2D
from fuseimg.data.ops.image_loader import OpLoadImage
from fuseimg.data.ops.color import OpClip, OpToRange


class OpSTOIC21SampleIDDecode(OpBase):
    """
    decodes sample id into image and segmentation filename
    """

    def __call__(self, sample_dict: NDict) -> NDict:
        """ """
        sid = get_sample_id(sample_dict)

        img_filename_key = "data.input.img_path"
        sample_dict[img_filename_key] = sid

        return sample_dict


class STOIC21:
    """
    Dataset created for COVID-19 AI challenge - https://stoic2021.grand-challenge.org/
    Aims to predict the severe outcome of COVID-19, based on the largest dataset of Computed Tomography (CT) images of COVID-19
    Each sample also include age, gender and targets rt-pcr result and outcome at 1 month: severe or non=severe. More details and download instructions can be found here: https://stoic2021.grand-challenge.org/stoic-db/.
    """

    # bump whenever the static pipeline modified
    STOIC21_DATASET_VER = 0

    @staticmethod
    def download(path: str) -> None:
        """
        Automatic download is not supported, please follow instructions in STOIC21 class header to download
        """
        assert (
            len(STOIC21.sample_ids(path)) > 0
        ), "automatic download is not supported, please follow instructions in STOIC21 class header to download"

    @staticmethod
    def sample_ids(path: str):
        """
        get all the sample ids in train-set
        sample_id is *.mha file found in the specified path
        """
        files = [os.path.join("data/mha/", os.path.basename(f)) for f in glob(os.path.join(path, "data/mha/*.mha"))]
        assert len(files) > 0, f"Expecting mha files in {os.path.join(path, 'data/mha/*.mha')}"
        return files

    @staticmethod
    def static_pipeline(data_path: str, output_shape: Tuple[int, int, int]) -> PipelineDefault:
        """
        Get suggested static pipeline (which will be cached), typically loading the data plus design choices that we won't experiment with.
        :param data_path: path to original STOIC21 data (See in STOIC21 header the instructions to download)
        :param output_shape: fixed shape to resize the image to
        """
        static_pipeline = PipelineDefault(
            "stoic21_static",
            [
                # decoding sample ID
                (
                    OpSTOIC21SampleIDDecode(),
                    dict(),
                ),  # will save image and seg path to "data.input.img_path", "data.gt.seg_path"
                # loading data
                (
                    OpLoadImage(data_path),
                    dict(key_in="data.input.img_path", key_out="data.input.img", key_metadata_out="data.metadata"),
                ),
                # resize
                # transposing so the depth channel will be first
                (
                    OpLambda(partial(np.moveaxis, source=-1, destination=0)),
                    dict(key="data.input.img"),
                ),  # convert image from shape [H, W, D] to shape [D, H, W]
                (
                    OpLambda(
                        partial(
                            skimage.transform.resize,
                            output_shape=output_shape,
                            mode="reflect",
                            anti_aliasing=True,
                            preserve_range=True,
                        )
                    ),
                    dict(key="data.input.img"),
                ),
                # read labels
                (OpToInt(), dict(key="data.metadata.PatientID")),
                (
                    OpReadDataframe(
                        data_filename=os.path.join(data_path, "metadata/reference.csv"),
                        key_column="data.gt.PatientID",
                        key_name="data.metadata.PatientID",
                        rename_columns={
                            "PatientID": "data.gt.PatientID",
                            "probCOVID": "data.gt.probCOVID",
                            "probSevere": "data.gt.probSevere",
                        },
                    ),
                    dict(),
                ),
            ],
        )
        return static_pipeline

    @staticmethod
    def dynamic_pipeline(train: bool, clip_range: Tuple[float, float]):
        """
        Get suggested dynamic pipeline. including pre-processing that might be modified and augmentation operations.
        :param clip_range: clip the original voxels values to fit this range
        """
        age_map = {"035Y": 0, "045Y": 1, "055Y": 2, "065Y": 3, "075Y": 4, "085Y": 5}
        gender_map = {"F": 0, "M": 1}
        dynamic_pipeline = PipelineDefault(
            "stoic21_dynamic",
            [
                # cast thickness to float
                (OpToFloat(), dict(key="data.metadata.SliceThickness")),
                # map input to categories
                (OpLookup(age_map), dict(key_in="data.metadata.PatientAge", key_out="data.input.age")),
                (OpToOneHot(len(age_map)), dict(key_in="data.input.age", key_out="data.input.age_one_hot")),
                (OpLookup(gender_map), dict(key_in="data.metadata.PatientSex", key_out="data.input.gender")),
                # create clinical data vector
                (
                    OpConcat(),
                    dict(
                        keys_in=["data.input.gender", "data.input.age_one_hot", "data.metadata.SliceThickness"],
                        key_out="data.input.clinical",
                    ),
                ),
                # fixed image normalization
                (OpToNumpy(), dict(key="data.input.img", dtype=np.float32)),  # cast to float
                (OpClip(), dict(key="data.input.img", clip=clip_range)),
                (OpToRange(), dict(key="data.input.img", from_range=clip_range, to_range=(0.0, 1.0))),
                # Numpy to tensor
                (OpToTensor(), dict(key="data.input.img", dtype=torch.float32)),
                (OpToTensor(), dict(key="data.input.clinical", dtype=torch.float32)),
                # add channel dimension -> [C=1, D, H, W]
                (OpLambda(partial(torch.unsqueeze, dim=0)), dict(key="data.input.img")),
            ],
        )

        # augmentation
        if train:
            dynamic_pipeline.extend(
                [
                    (OpLambda(partial(torch.squeeze, dim=0)), dict(key="data.input.img")),
                    # affine augmentation - will apply the same affine transformation on each slice
                    (
                        OpRandApply(OpSample(OpAugAffine2D()), 0.5),
                        dict(
                            key="data.input.img",
                            rotate=Uniform(-180.0, 180.0),
                            scale=Uniform(0.8, 1.2),
                            flip=(RandBool(0.5), RandBool(0.5)),
                            translate=(RandInt(-15, 15), RandInt(-15, 15)),
                        ),
                    ),
                    # color augmentation - check if it is useful in CT images
                    # (OpSample(OpAugColor()), dict(
                    #     key="data.input.img",
                    #     gamma=Uniform(0.8,1.2),
                    #     contrast=Uniform(0.9,1.1),
                    #     add=Uniform(-0.01, 0.01)
                    # )),
                    # add channel dimension -> [C=1, D, H, W]
                    (OpLambda(partial(torch.unsqueeze, dim=0)), dict(key="data.input.img")),
                ]
            )

        return dynamic_pipeline

    @staticmethod
    def dataset(
        data_path: str,
        cache_dir: str,
        reset_cache: bool = False,
        num_workers: int = 10,
        sample_ids: Optional[Sequence[Hashable]] = None,
        train: bool = False,
        output_shape: Tuple[int, int, int] = (32, 256, 256),
        clip_range: Tuple[float, float] = (-200, 800),
    ) -> DatasetDefault:
        """
        Get cached dataset
        :param data_path: path to store the original data
        :param cache_dir: path to store the cache
        :param reset_cache: set to True tp reset the cache
        :param num_workers: number of processes used for caching
        :param sample_ids: dataset including the specified sample_ids or None for all the samples. sample_id is case_{id:05d} (for example case_00001 or case_00100).
        :param train: True if used for training  - adds augmentation operations to the pipeline
        :param output_shape: fixed shape to resize the image to
        :param clip_range: clip the original voxels values to fit this range
        """

        if sample_ids is None:
            sample_ids = STOIC21.sample_ids(data_path)

        static_pipeline = STOIC21.static_pipeline(data_path, output_shape=output_shape)
        dynamic_pipeline = STOIC21.dynamic_pipeline(train=train, clip_range=clip_range)

        cacher = SamplesCacher(
            f"stoic_cache_ver{STOIC21.STOIC21_DATASET_VER}",
            static_pipeline,
            [cache_dir],
            restart_cache=reset_cache,
            workers=num_workers,
        )

        my_dataset = DatasetDefault(
            sample_ids=sample_ids,
            static_pipeline=static_pipeline,
            dynamic_pipeline=dynamic_pipeline,
            cacher=cacher,
        )
        my_dataset.create()
        return my_dataset



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Union

import numpy as np
import torch

from fuse.dl.losses.loss_base import LossBase
from fuse.utils.ndict import NDict


class LossSegmentationCrossEntropy(LossBase):
    def __init__(
        self,
        pred_name: str = None,
        target_name: str = None,
        weight: float = 1.0,
        class_weights: Union[str, torch.tensor, None] = "ipw",
        resize_mode: str = "interpolate",
    ) -> None:
        """
        SegmentationCrossEntropy loss.
        This class calculates cross entropy loss per location ("dense") of a class activation map ("segmentation"),
        or more correctly, logits of a class activation map. Ground truth map should be a float32 dtype, with binary
        entries. It is resized (reduced) to match dimensions of predicted score map.
        Note: The cross entropy function is weighted with inverse proportion of positive pixels.
        :param pred_name:               batch_dict key for prediction (e.g., network output)
        :param target_name:             batch_dict key for target (e.g., ground truth label)
        :param weight:                  Weight multiplier for final loss value
        :param class_weights:           Class weights. Options:
                                            1. None
                                            2. Tensor with class weights
                                            3. 'ipw' = Inverse proportional weighting
        :param resize_mode:             Resize mode- either using a max pooling kernel, or using PyTorch interpolation (default)
        """

        self.pred_name = pred_name
        self.target_name = target_name
        self.weight = weight
        self.class_weights = class_weights
        self.resize_mode = resize_mode
        assert resize_mode in ["maxpool", "interpolate"]

    def __call__(self, batch_dict: NDict) -> torch.Tensor:
        preds = batch_dict[self.pred_name]  # preds shape [batch_size, num_classes, height, width]
        targets = batch_dict[self.target_name]  # targets shape [batch_size, height, width]

        batch_size, targets_height, targets_width = targets.shape
        batch_size, num_classes, preds_height, preds_width = preds.shape

        # Resize targets (if needed) to match predicted activation map
        # ============================================================
        targets = targets.reshape([batch_size, 1, targets_height, targets_width])

        if (targets_height > preds_height) or (targets_width > preds_width):
            if targets.dtype != torch.float32:
                targets = targets.type(torch.float32).to(targets.device)
            if self.resize_mode == "maxpool":
                block_height = int(targets_height / preds_height)
                block_width = int(targets_width / preds_width)
                residual_h = int((targets_height - (block_height * preds_height)) / 2)
                residual_w = int((targets_width - (block_width * preds_width)) / 2)

                targets = torch.nn.functional.max_pool2d(
                    targets[:, :, residual_h : targets_height - residual_h, residual_w : targets_width - residual_w],
                    kernel_size=(block_height, block_width),
                )
            elif self.resize_mode == "interpolate":
                targets = torch.nn.functional.interpolate(targets, size=(preds_height, preds_width))
            else:
                raise Exception

        # Generate class weights
        # ======================
        if self.class_weights is None:
            segmentation_weights = None
        elif isinstance(self.class_weights, str) and self.class_weights.lower() == "ipw":
            # =============================================================================
            # NOTE: PyTorch also has an equivalent 'unique' method with counts, but it was introduced in a later version
            # See: https://github.com/pytorch/pytorch/issues/12598
            values, counts = np.unique(targets.cpu(), return_counts=True)
            # =============================================================================
            values = values.astype("uint16")
            segmentation_weights = np.zeros(num_classes, dtype="float32")
            segmentation_weights[values] = 1 / counts
            segmentation_weights = segmentation_weights * (1 / segmentation_weights.max())
            segmentation_weights = torch.FloatTensor(segmentation_weights).to(targets.device)
        elif isinstance(self.class_weights, (torch.tensor, list, tuple)):
            segmentation_weights = torch.FloatTensor(self.class_weights).to(targets.device)

        # Reshape targets and change its dtype to int64 for cross entropy input
        # =====================================================================
        targets = targets.reshape([batch_size, preds_height, preds_width])
        if targets.dtype != torch.int64:
            targets = targets.type(torch.int64).to(targets.device)

        # Calc loss
        # ==========
        loss_obj = torch.nn.functional.cross_entropy(preds, targets, weight=segmentation_weights) * self.weight

        return loss_obj
